""" Main Script from which the whole program runs.
"""
from argparse import ArgumentParser
import os
import warnings
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' # pylint: disable=wrong-import-position

import pandas as pd
from scipy.stats import ConstantInputWarning
import tensorflow as tf

from inspire.calibration import calibrate
from inspire.config import Config
from inspire.convert import convert_raw_to_mgf
from inspire.constants import ENDC_TEXT, OKGREEN_TEXT
from inspire.download import download_data, download_models, download_utils
from inspire.pepseek.extract_candidates import extract_epitope_candidates
from inspire.execute_msfragger import execute_msfragger
from inspire.get_spectral_angle import get_spectral_angle
from inspire.input.search_results import generic_read_df
from inspire.plot_spectra.plot_isobars import plot_isobars
from inspire.plot_spectra.plot_spectra import plot_spectra
from inspire.predict_binding import predict_binding
from inspire.predict_spectra import predict_spectra
from inspire.prepare import prepare_for_spectral_prediction, prepare_for_mhcpan
from inspire.feature_creation import create_features
from inspire.feature_selection import select_features
from inspire.quant.execute import quantify_identifications
from inspire.quant.normalise import normalise_intensities
from inspire.quant.de_analysis import de_analysis
from inspire.quant.report_template import create_quant_report
from inspire.rescore import final_rescoring
from inspire.report import generate_report
from inspire.utils import fetch_collision_energy
from inspire.validate import validate_spliced
import inspire

pd.options.mode.chained_assignment = None
tf.config.set_visible_devices([], 'GPU')
warnings.filterwarnings("ignore", category=ConstantInputWarning)


PIPELINE_OPTIONS = [
    'calibrate',
    'core',
    'convert',
    'downloadExample',
    'format',
    'fragger',
    'extractCandidates',
    'featureGeneration',
    'featureSelection',
    'featureSelection+',
    'finalRescoring',
    'generateReport',
    'panPrepare',
    'plotIsobars',
    'plotSpectra',
    'predictBinding',
    'predictSpectra',
    'prepare',
    'quantify',
    'rescore',
    'spectralAngle',
    'spectralPrepare',
]

def get_arguments():
    """ Function to collect command line arguments.

    Returns
    -------
    args : argparse.Namespace
        The parsed command line arguments.
    """
    parser = ArgumentParser(description='inSPIRE Pipeline for MS Search Results.')

    parser.add_argument(
        '--config_file',
        help='Config file to be read from.'
    )

    parser.add_argument(
        '--pipeline',
        choices=PIPELINE_OPTIONS,
        required=True,
        help='What pipeline do you want to run?',
    )

    return parser.parse_args()

def run_inspire(pipeline=None, config_file=None):
    """ Function to orchestrate running of the whole ininspire package.
    """
    print(f'\n---> Running inSPIRE version {inspire.__version__} <---\n')
    if pipeline is None:
        args = get_arguments()
        config_file = args.config_file
        pipeline = args.pipeline

    if pipeline == 'downloadExample':
        download_data()
    else:
        config = Config(config_file)
        config.validate()
        print(
            OKGREEN_TEXT +
            'Checking for required inSPIRE models...' +
            ENDC_TEXT
        )
        download_models(force_reload=config.force_reload)
        download_utils(force_reload=config.force_reload)

    if pipeline == 'calibrate' or (
        config.collision_energy is None and
        not os.path.exists(f'{config.output_folder}/collisionEnergyStats.csv')
        and pipeline not in ('convert', 'fragger')
    ):
        print(
            OKGREEN_TEXT +
            'Running CE Calibration...' +
            ENDC_TEXT
        )
        calibrate(config)

    if config.collision_energy is None and pipeline not in ('convert', 'fragger'):
        config.collision_energy = fetch_collision_energy(config.output_folder)

    if pipeline == 'convert':
        print(
            OKGREEN_TEXT +
            'Creating Formatted Spectral Prediction Input...' +
            ENDC_TEXT
        )
        convert_raw_to_mgf(config)

    if pipeline == 'fragger':
        print(
            OKGREEN_TEXT +
            'Executing MSFragger with default inSPIRE settings...' +
            ENDC_TEXT
        )
        execute_msfragger(config)

    if pipeline == 'format':
        print(
            OKGREEN_TEXT +
            'Formatting search results for inSPIRE input...' +
            ENDC_TEXT
        )
        _ = generic_read_df(config)

    if pipeline in ('spectralPrepare', 'prepare', 'core'):
        print(
            OKGREEN_TEXT +
            'Creating Formatted Spectral Prediction Input...' +
            ENDC_TEXT
        )
        prepare_for_spectral_prediction(config)

    if pipeline in ('panPrepare', 'prepare', 'core'):
        if config.use_binding_affinity is not None:
            print(
                OKGREEN_TEXT +
                'Creating Formatted NetMHCpan Input...' +
                ENDC_TEXT
            )
        prepare_for_mhcpan(config)

    if pipeline in ('predictSpectra', 'core'):
        print(
            OKGREEN_TEXT +
            'Predicting Spectra...' +
            ENDC_TEXT
        )
        predict_spectra(config, 'core')

    if pipeline in ('predictBinding', 'core'):
        if config.use_binding_affinity is not None:
            print(
                OKGREEN_TEXT +
                'Predicting NetMHCpan Binding Affinity...' +
                ENDC_TEXT
            )
            predict_binding(config)

    if pipeline in ('featureGeneration', 'rescore', 'core'):
        print(
            OKGREEN_TEXT +
            'Generating Features for Percolator Input...' +
            ENDC_TEXT
        )
        create_features(config)

    if pipeline in ('featureSelection', 'featureSelection+', 'rescore', 'core'):
        print(
            OKGREEN_TEXT +
            'Optimising Feature Set...' +
            ENDC_TEXT
        )
        select_features(config)

    if pipeline in ('finalRescoring', 'featureSelection+', 'rescore', 'core'):
        print(
            OKGREEN_TEXT +
            'Running Finalised Rescoring...' +
            ENDC_TEXT
        )
        final_rescoring(config)

    if (
        pipeline in ('validate', 'featureSelection+', 'rescore', 'core', 'calibrate+core')
        and config.accession_format == 'invitroSPI'
    ):
        print(
            OKGREEN_TEXT +
            'Validating spliced assignments...' +
            ENDC_TEXT
        )
        validate_spliced(config)

    if (
        pipeline in ('featureSelection+', 'rescore', 'core') and not config.silent_execution
    ) or pipeline == 'generateReport':
        print(
            OKGREEN_TEXT +
            'Generating inSPIRE Performance Report...' +
            ENDC_TEXT
        )
        generate_report(config)

    if (
        pipeline in ('validate', 'featureSelection+', 'rescore', 'core', 'calibrate+core')
        and config.use_accession_stratum
    ):
        print(
            OKGREEN_TEXT +
            'Validating spliced assignments...' +
            ENDC_TEXT
        )
        validate_spliced(config)

    if pipeline == 'spectralAngle':
        print(
            OKGREEN_TEXT +
            'Calculating Spectral Angles...' +
            ENDC_TEXT
        )
        get_spectral_angle(config)

    if pipeline == 'quantify':
        print(
            OKGREEN_TEXT +
            'Running quantification via skyline docker...' +
            ENDC_TEXT
        )
        quantify_identifications(config)
        normalise_intensities(config)
        de_analysis(config)
        create_quant_report(config)


    if pipeline == 'extractCandidates':
        print(
            OKGREEN_TEXT +
            'Extracting Potential Epitope Candidates...' +
            ENDC_TEXT
        )
        extract_epitope_candidates(config)

    if pipeline == 'plotSpectra':
        print(
            OKGREEN_TEXT +
            'Plotting Spectra...' +
            ENDC_TEXT
        )
        plot_spectra(config)

    if pipeline == 'plotIsobars':
        print(
            OKGREEN_TEXT +
            'Plotting Isobars...' +
            ENDC_TEXT
        )
        plot_isobars(config)

    print(
        OKGREEN_TEXT +
        f'inSPIRE Pipeline "{pipeline}" Complete!' +
        ENDC_TEXT
    )

if __name__ == '__main__':
    run_inspire()



""" Function for downloading the models required for inSPIRE execution.
"""
import os
from pathlib import Path
from urllib.request import urlretrieve
import shutil
import tarfile
import zipfile

from inspire.constants import (
    ENDC_TEXT,
    FIGSHARE_EXAMPLE_PATH,
    FIGSHARE_EXTERNAL_UTILS_PATH,
    FIGSHARE_PATH,
    OKCYAN_TEXT,
    THERMO_PARSER_PATH,
)


def download_thermo_raw_file_parser():
    """ Function to download the ThermoRawFileParser.
    """
    home = str(Path.home())
    if os.path.isdir(f'{home}/inSPIRE_models/ThermoRawFileParser'):
        print(
            OKCYAN_TEXT + '\tThermoRawFileParser already downloaded.' + ENDC_TEXT
        )
    else:
        os.mkdir(f'{home}/inSPIRE_models/ThermoRawFileParser')
        print(
            OKCYAN_TEXT + '\tDownloading ThermoRawFileParser...' + ENDC_TEXT
        )
        urlretrieve(THERMO_PARSER_PATH, f'{home}/inSPIRE_models/ThermoRawFileParser/parser.zip')
        print(
            OKCYAN_TEXT + '\tExtracting ThermoRawFileParser...' + ENDC_TEXT
        )
        with zipfile.ZipFile(f'{home}/inSPIRE_models/ThermoRawFileParser/parser.zip') as zip_ref:
            zip_ref.extractall(f'{home}/inSPIRE_models/ThermoRawFileParser')
        print(
            OKCYAN_TEXT + '\tThermoParserReady ready.' + ENDC_TEXT
        )


def download_models(force_reload=False):
    """ Function to download the required models for inSPIRE execution from
        figshare.

    Parameters
    ----------
    force_reload : bool (default=False)
        Flag indicating whether to remove the existing inSPIRE_models folder
        and redownload all models.
    """
    home = str(Path.home())
    if force_reload:
        os.rmdir(f'{home}/inSPIRE_models')

    if os.path.isdir(f'{home}/inSPIRE_models/models'):
        print(
            OKCYAN_TEXT + '\tModels already downloaded.' + ENDC_TEXT
        )
    else:
        if not os.path.isdir(f'{home}/inSPIRE_models'):
            os.mkdir(f'{home}/inSPIRE_models')
        print(
            OKCYAN_TEXT + '\tDownloading models...' + ENDC_TEXT
        )
        urlretrieve(FIGSHARE_PATH, f'{home}/inSPIRE_models/models.zip')
        print(
            OKCYAN_TEXT + '\tExtracting Models...' + ENDC_TEXT
        )
        with zipfile.ZipFile(f'{home}/inSPIRE_models/models.zip') as zip_ref:
            zip_ref.extractall(f'{home}/inSPIRE_models/models')
        print(
            OKCYAN_TEXT + '\tModels ready.' + ENDC_TEXT
        )

def download_utils(force_reload=False):
    """ Function to download the required models for inSPIRE execution from
        figshare.

    Parameters
    ----------
    force_reload : bool (default=False)
        Flag indicating whether to remove the existing inSPIRE_models folder
        and redownload all models.
    """
    home = str(Path.home())
    if force_reload:
        shutil.rmtree(f'{home}/inSPIRE_models/utilities')

    download=True
    if os.path.isfile(f'{home}/inSPIRE_models/utilities/version.txt'):
        with open(f'{home}/inSPIRE_models/utilities/version.txt', 'r') as version_file:
            version = version_file.read().strip()
            if version == '>=2.0':
                print(
                    OKCYAN_TEXT + '\tUtils already downloaded.' + ENDC_TEXT
                )
                download=False

    if download:
        if os.path.isdir(f'{home}/inSPIRE_models/utilities'):
            shutil.rmtree(f'{home}/inSPIRE_models/utilities')

        os.mkdir(f'{home}/inSPIRE_models/utilities')
        print(
            OKCYAN_TEXT + '\tDownloading external utilities...' + ENDC_TEXT
        )
        urlretrieve(FIGSHARE_EXTERNAL_UTILS_PATH, f'{home}/inSPIRE_models/utilities/utils.zip')
        print(
            OKCYAN_TEXT + '\tExtracting utils...' + ENDC_TEXT
        )
        with zipfile.ZipFile(f'{home}/inSPIRE_models/utilities/utils.zip') as zip_ref:
            zip_ref.extractall(f'{home}/inSPIRE_models/utilities')
        print(
            OKCYAN_TEXT + '\tUtils ready.' + ENDC_TEXT
        )


def download_data():
    """ Function to download the example dataset from Figshare

    Parameters
    ----------
    force_reload : bool (default=False)
        Flag indicating whether to remove the existing inSPIRE_models folder
        and redownload all models.
    """

    if os.path.isdir('example'):
        print(
            OKCYAN_TEXT + '\tExample data already downloaded.' + ENDC_TEXT
        )
    else:
        print(
            OKCYAN_TEXT + '\tDownloading data...' + ENDC_TEXT
        )
        urlretrieve(FIGSHARE_EXAMPLE_PATH, filename=f'{os.getcwd()}/example.tar.gz')
        print(
            OKCYAN_TEXT + '\tExtracting Data...' + ENDC_TEXT
        )
        with tarfile.open('example.tar.gz', "r:gz") as tar:
            tar.extractall()
        print(
            OKCYAN_TEXT + '\tDataset ready.' + ENDC_TEXT
        )



""" Code for running Prosit models on CPU.
    Large sections based on https://github.com/kusterlab/prosit.
"""
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import regularizers, constraints, initializers
from tensorflow.keras.layers import Layer # pylint: disable=no-name-in-module
import yaml

from inspire.constants import (
    MAX_CHARGE,
    MAX_FRAG_CHARGE,
    MAX_SEQ_LEN,
    N_ION_TYPES,
    N_LOSSES,
    PROSIT_ALPHABET,
    PROSIT_MASK_VALUE,
    PROSIT_PRED_BATCH_SIZE,
    PROSIT_UNMOD_ALPHA_S,
)

PROSIT_IONS = np.array(
    [
        'y1', 'y1^2)', 'y1^3)', 'b1', 'b1^2)', 'b1^3)',
        'y2', 'y2^2)', 'y2^3)', 'b2', 'b2^2)', 'b2^3)',
        'y3', 'y3^2)', 'y3^3)', 'b3', 'b3^2)', 'b3^3)',
        'y4', 'y4^2)', 'y4^3)', 'b4', 'b4^2)', 'b4^3)',
        'y5', 'y5^2)', 'y5^3)', 'b5', 'b5^2)', 'b5^3)',
        'y6', 'y6^2)', 'y6^3)', 'b6', 'b6^2)', 'b6^3)',
        'y7', 'y7^2)', 'y7^3)', 'b7', 'b7^2)', 'b7^3)',
        'y8', 'y8^2)', 'y8^3)', 'b8', 'b8^2)', 'b8^3)',
        'y9', 'y9^2)', 'y9^3)', 'b9', 'b9^2)', 'b9^3)',
        'y10', 'y10^2', 'y10^3', 'b10', 'b10^2', 'b10^3',
        'y11', 'y11^2', 'y11^3', 'b11', 'b11^2', 'b11^3',
        'y12', 'y12^2', 'y12^3', 'b12', 'b12^2', 'b12^3',
        'y13', 'y13^2', 'y13^3', 'b13', 'b13^2', 'b13^3',
        'y14', 'y14^2', 'y14^3', 'b14', 'b14^2', 'b14^3',
        'y15', 'y15^2', 'y15^3', 'b15', 'b15^2', 'b15^3',
        'y16', 'y16^2', 'y16^3', 'b16', 'b16^2', 'b16^3',
        'y17', 'y17^2', 'y17^3', 'b17', 'b17^2', 'b17^3',
        'y18', 'y18^2', 'y18^3', 'b18', 'b18^2', 'b18^3',
        'y19', 'y19^2', 'y19^3', 'b19', 'b19^2', 'b19^3',
        'y20', 'y20^2', 'y20^3', 'b20', 'b20^2', 'b20^3',
        'y21', 'y21^2', 'y21^3', 'b21', 'b21^2', 'b21^3',
        'y22', 'y22^2', 'y22^3', 'b22', 'b22^2', 'b22^3',
        'y23', 'y23^2', 'y23^3', 'b23', 'b23^2', 'b23^3',
        'y24', 'y24^2', 'y24^3', 'b24', 'b24^2', 'b24^3',
        'y25', 'y25^2', 'y25^3', 'b25', 'b25^2', 'b25^3',
        'y26', 'y26^2', 'y26^3', 'b26', 'b26^2', 'b26^3',
        'y27', 'y27^2', 'y27^3', 'b27', 'b27^2', 'b27^3',
        'y28', 'y28^2', 'y28^3', 'b28', 'b28^2', 'b28^3',
        'y29', 'y29^2', 'y29^3', 'b29', 'b29^2', 'b29^3',
    ],
    dtype='str',
)

class PrositAttention(Layer):
    """ The modified attention layer used by Prosit.
    """
    def __init__(
        self,
        context=False,
        w_regularizer=None,
        b_regularizer=None,
        u_regularizer=None,
        w_constraint=None,
        b_constraint=None,
        u_constraint=None,
        bias=True,
        **kwargs
    ):
        self.supports_masking = True
        self.init = initializers.get("glorot_uniform")
        self.w_regularizer = regularizers.get(w_regularizer)
        self.b_regularizer = regularizers.get(b_regularizer)
        self.u_regularizer = regularizers.get(u_regularizer)
        self.w_constraint = constraints.get(w_constraint)
        self.b_constraint = constraints.get(b_constraint)
        self.u_constraint = constraints.get(u_constraint)
        self.bias = bias
        self.context = context
        self.b = None
        self.w = None
        self.u = None
        super().__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 3
        self.w = self.add_weight(
            shape=(input_shape[-1],),
            initializer=self.init,
            name=f'{self.name}_W',
            regularizer=self.w_regularizer,
            constraint=self.w_constraint,
        )
        if self.bias:
            self.b = self.add_weight(
                shape=(input_shape[1],),
                initializer="zero",
                name=f'{self.name}_b',
                regularizer=self.b_regularizer,
                constraint=self.b_constraint,
            )
        else:
            self.b = None
        if self.context:
            self.u = self.add_weight(
                shape=(input_shape[-1],),
                initializer=self.init,
                name=f'{self.name}_u',
                regularizer=self.u_regularizer,
                constraint=self.u_constraint,
            )

        self.built = True

    def call(self, inputs, *args, **kwargs):
        """ Function to execute the PrositAttention layer.
        """
        processed_tensor = K.squeeze(K.dot(inputs, K.expand_dims(self.w)), axis=-1)
        if self.bias:
            processed_tensor += self.b
        processed_tensor = K.tanh(processed_tensor)
        if self.context:
            processed_tensor = K.squeeze(K.dot(inputs, K.expand_dims(self.u)), axis=-1)
        processed_tensor = K.exp(processed_tensor)

        processed_tensor /= K.cast(
            K.sum(processed_tensor, axis=1, keepdims=True) + K.epsilon(), K.floatx()
        )
        processed_tensor = K.expand_dims(processed_tensor)
        weighted_input = inputs * processed_tensor

        return K.sum(weighted_input, axis=1)

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[-1]

    def get_config(self):
        config = {
            "bias": self.bias,
            "context": self.context,
            "w_regularizer": regularizers.serialize(self.w_regularizer),
            "b_regularizer": regularizers.serialize(self.b_regularizer),
            "u_regularizer": regularizers.serialize(self.u_regularizer),
            "w_constraint": constraints.serialize(self.w_constraint),
            "b_constraint": constraints.serialize(self.b_constraint),
            "u_constraint": constraints.serialize(self.u_constraint),
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

def load_model(config_loc, model_loc, weights_loc):
    """ Function to load a Prosit model from disk.

    Parameters
    ----------
    config_loc : str
        The location of the model config file.
    model_loc : str
        The location of the model file.
    weights_loc : str
        The location of the model weights file.

    Returns
    -------
    model_dict : dict
        The loaded model dictionary.
    """
    with open(config_loc, 'r', encoding='UTF-8') as stream:
        config_dict = yaml.safe_load(stream)

    with open(model_loc, 'r', encoding='UTF-8') as model_file:
        model = tf.keras.models.model_from_json(
            model_file.read(), custom_objects={"PrositAttention": PrositAttention}
        )

    model.load_weights(weights_loc)

    model_dict = {}
    model_dict["graph"] = tf.Graph()

    model_dict["session"] = tf.compat.v1.Session()
    with model_dict["session"].as_default():
        model_dict["model"], model_dict["config"] = model, config_dict
        model_dict["model"].compile(optimizer="adam", loss="mse")

    return model_dict

def get_precursor_charge_onehot(charges):
    """ Function to get one hot array from precursor charge.

    Parameters
    ----------
    charges : np.array of int
        The charge states observed.

    Return
    ------
    charge_one_hot : np.array of np.array
        Array of the one hot encoded charge states.
    """
    charge_one_hot = np.zeros([len(charges), MAX_CHARGE], dtype=int)
    for i, precursor_charge in enumerate(charges):
        charge_one_hot[i, precursor_charge - 1] = 1
    return charge_one_hot


def peptide_parser(peptide):
    """ Function to yield residues possible with modification.

    Parameters
    ----------
    peptide : str
        The input peptide.

    Yields
    ------
    residue : str
        Either a single character residue or M(ox).
    """
    pep_len = len(peptide)
    i = 0
    while i < pep_len:
        if i < pep_len - 3 and peptide[i + 1] == '(':
            j = peptide[i + 2 :].index(')')
            offset = i + j + 3
            yield PROSIT_ALPHABET[peptide[i:offset]]
            i = offset
        else:
            yield PROSIT_ALPHABET[peptide[i]]
            i += 1

def get_sequence_integer(sequences):
    """ Function to create np arrays of integers from peptides using the Prosit alphabet.

    Parameters
    ----------
    sequences : list of str
        The peptide sequences.

    Returns
    -------
    sequence_integers : np.array
        The encoded arrays.
    """
    sequence_integers = np.zeros([len(sequences), MAX_SEQ_LEN], dtype=int)
    for seq_idx, sequence in enumerate(sequences):
        for res_idx, residue in enumerate(peptide_parser(sequence)):
            sequence_integers[seq_idx, res_idx] = residue
    return sequence_integers


def sanitize(data):
    """ Function to sanitize Prosit predicted MS2 spectra.

    Parameters
    ----------
    data : dict
        The output of the Prosit model.

    Returns
    -------
    data : dict
        The Prosit output after cleaning and normalising the spectrum.
    """
    sequence_lengths = np.count_nonzero(data["sequence_integer"], axis=1)
    intensities = data["intensities_pred"]
    charges = list(data["precursor_charge_onehot"].argmax(axis=1) + 1)

    intensities[intensities < 0] = 0
    maxima = intensities.max(axis=1)
    intensities /= maxima[:, np.newaxis]

    intensities = intensities.reshape(
        [intensities.shape[0], MAX_SEQ_LEN - 1, N_ION_TYPES, N_LOSSES, MAX_FRAG_CHARGE]
    )

    for i in range(intensities.shape[0]):
        intensities[i, sequence_lengths[i] - 1 :, :, :, :] = PROSIT_MASK_VALUE
    for i in range(intensities.shape[0]):
        if charges[i] < 3:
            intensities[i, :, :, :, charges[i] :] = PROSIT_MASK_VALUE

    flat_dim = [
        intensities.shape[0],
        (MAX_SEQ_LEN - 1)*N_ION_TYPES*N_LOSSES*MAX_FRAG_CHARGE,
    ]
    intensities = intensities.reshape(flat_dim)
    data["intensities_pred"] = intensities

    return data


def prosit_predict(data, d_model):
    """ Function to predict MS2 spectra or iRT using a Prosit model.

    Parameters
    ----------
    data : dict
        A dictionary containing modified_sequence and precursor_charge keys as well as
        collision_energy if MS2 spectrum is being predicted.
    d_model : dict
        A dictionary containing the model details.

    Return
    ------
    data : dict
        The input dictionary updated to contain the prediction data.
    """
    # check for mandatory keys
    input_data = [data[key] for key in d_model["config"]["x"]]

    prediction = d_model["model"].predict(
        input_data, verbose=True, batch_size=PROSIT_PRED_BATCH_SIZE
    )

    if d_model["config"]["prediction_type"] == "intensity":
        data["intensities_pred"] = prediction
        data = sanitize(data)
    elif d_model["config"]["prediction_type"] == "iRT":
        scal = float(d_model["config"]["iRT_rescaling_var"])
        mean = float(d_model["config"]["iRT_rescaling_mean"])
        data["iRT"] = prediction * np.sqrt(scal) + mean
    else:
        raise ValueError("model_config misses parameter")

    return data


def generate_mods_string_tuples(sequence_integer):
    """ Function to generate tuples of the modifications in a sequence.

    Parameters
    ----------
    sequence_integer : np.array of int
        The peptide sequence encoded using the Prosit alphabet.

    Returns
    -------
    list_mods : list of tuples
        A list of the modifications stored as tuples containing location and identity.
    """
    list_mods = []
    for mod in [PROSIT_ALPHABET['M(ox)'], PROSIT_ALPHABET['C']]:
        for position in np.where(sequence_integer == mod)[0]:
            if mod == PROSIT_ALPHABET['C']:
                list_mods.append((position + 1, "C", "Carbamidomethyl"))
            elif mod == PROSIT_ALPHABET['M(ox)']:
                list_mods.append((position + 1, "M", "Oxidation"))
            else:
                raise ValueError(f'Modification ID {mod} not allowed.')
    list_mods.sort(key=lambda tup: tup[0])

    return list_mods


def generate_mod_strings(sequence_integer):
    """
    >>> x = np.array([1,2,3,1,2,21,0])
    >>> y, z = generate_mod_strings(x)
    >>> y
    '3/1,C,Carbamidomethyl/4,C,Carbamidomethyl/5,M,Oxidation'
    >>> z
    'Carbamidomethyl@C2; Carbamidomethyl@C5; Oxidation@M6'
    """
    list_mods = generate_mods_string_tuples(sequence_integer)
    if len(list_mods) == 0:
        return '0', ''
    indexed_mods_string = ''
    at_mods_string = ''
    indexed_mods_string += str(len(list_mods))
    for i, mod_tuple in enumerate(list_mods):
        indexed_mods_string += (
            '/' + str(mod_tuple[0] + 1) + ',' + mod_tuple[1] + ',' + mod_tuple[2]
        )
        if i == 0:
            at_mods_string += (
                mod_tuple[2] + '@' + mod_tuple[1] + str(mod_tuple[0] + 1)
            )
        else:
            at_mods_string += (
                '; ' + mod_tuple[2] + '@' + mod_tuple[1] + str(mod_tuple[0] + 1)
            )

    return indexed_mods_string, at_mods_string


def format_msp_spectrum(
        pred_intes,
        collision_energy,
        pred_irt,
        precursor_charge,
        sequence_integer,
        pred_ion_names,
    ):
    """ Function to correctly format an MSP spectrum from Prosit predictions.

    Parameters
    ----------
    pred_intes : np.array
        The predicted intensities from Prosit.
    collision_energy : int
        The collision energy predicted for.
    pred_irt : float
        The predicted iRT value from Prosit.
    precursor_charge : int
        The charge of the peptide.
    sequence_integer : np.array
        The encoded peptide sequence.
    pred_ion_names : np.array
        The name of the ions matching the intensities.
    """
    mod, mod_string = generate_mod_strings(sequence_integer)
    unmod_seq = ''.join(
        [PROSIT_UNMOD_ALPHA_S[i] if i in PROSIT_UNMOD_ALPHA_S else '' for i in sequence_integer]
    )

    print_string = (
        f'Name: {unmod_seq}/{precursor_charge}\nMW: 0.0\n' +
        f'Comment: Parent=0.0 Collision_energy={np.round(collision_energy, 0)} ' +
        f'Mods={mod} ModString={unmod_seq}//{mod_string}/{precursor_charge}' +
        f' iRT={pred_irt}' +
        f'\nNum peaks: {len(pred_intes)}'
    )

    for intensity, ion in zip(pred_intes, pred_ion_names):
        print_string += '\n0.0\t' + str(intensity) + '\t"'
        print_string += ion + '/0.0ppm"'

    return print_string


def write_msp_spectrum(df_row, out_file, chunk_idx):
    """ Function to write a single spectrum in MSP format.

    Parameters
    ----------
    df_row : pd.Series
        A single row of the DataFrame containing prediction data.
    out_file : file
        The File where the spectrum will be written.
    chunk_idx : str
        The index of the chunk being predicted for.
    """
    first_spec = df_row.name == 0 and chunk_idx == 0
    pred_intes = df_row['intensities_pred']
    sel = np.where(pred_intes > 0)
    pred_intes = pred_intes[sel]
    collision_energy = df_row['collision_energy']
    pred_irt = df_row['iRT']
    precursor_charge = df_row['precursor_charge']
    sequence_integer = df_row['sequence_integer']
    pred_ion_names = PROSIT_IONS[sel]
    spec = format_msp_spectrum(
        pred_intes,
        collision_energy,
        pred_irt,
        precursor_charge,
        sequence_integer,
        pred_ion_names,
    )
    if not first_spec:
        out_file.write("\n")
    first_spec = False
    out_file.write(str(spec))


def write_msp_file(peptide_df, prediction_data, out_path, chunk_idx):
    """ Function to write MSP output from Prosit predictions.

    Parameters
    ----------
    peptide_df : pd.DataFrame
        The input DataFrame used  from prosit prediction.
    prediction_data : dict
        Dictionary of the Prosit output predictions.
    out_path : str
        The location where results should be written.
    """
    peptide_df['iRT'] = prediction_data['iRT']
    peptide_df['intensities_pred'] = pd.Series(list(prediction_data['intensities_pred']))
    peptide_df['sequence_integer'] = pd.Series(list(prediction_data['sequence_integer']))
    if chunk_idx > 0:
        write_mode = 'a'
    else:
        write_mode = 'w'
    with open(out_path, mode=write_mode, encoding='UTF-8') as out_file:
        peptide_df.apply(lambda df_row : write_msp_spectrum(df_row, out_file, chunk_idx), axis=1)



""" Functions for executing MSFragger Searches from within inSPIRE.
"""
import os
from pathlib import Path
import shutil
import sys


UNSPECIFIC_CLEAVAGE = (
    'search_enzyme_name_1 = nonspecific\n'+
    'search_enzyme_cut_1 = -\n' +
    'search_enzyme_nocut_1 = \n' +
    'search_enzyme_sense_1 = C\n' +
    'allowed_missed_cleavage_1 = 2\n'
)

TRYPTIC_CLEAVAGE = (
    'search_enzyme_name_1 = trypsin\n'+
    'search_enzyme_cut_1 = KR\n' +
    'search_enzyme_nocut_1 = \n' +
    'search_enzyme_sense_1 = C\n' +
    'allowed_missed_cleavage_1 = 2\n'
)

def get_proteins(protein_file, all_ids, all_proteins):
    """ Function to retrieve protein IDs and names from a file.

    Parameters
    ----------
    protein_file : str
        Path to a protein fasta file.
    all_ids : list of str
        A list of the IDs of the proteins in the file.
    all_proteins : list of str
        A list of the sequences of the proteins in the file.

    Returns
    -------
    all_ids : list of str
        A list of the IDs of the proteins in the file.
    all_proteins : list of str
        A list of the sequences of the proteins in the file.
    """
    with open(protein_file, mode='r', encoding='UTF-8') as fasta_f:
        protein = ''
        while (line := fasta_f.readline()):
            if line.startswith('>'):
                all_ids.append(line[1:].strip('\n'))
                if protein:
                    all_proteins.append(protein)
                    protein = ''
            else:
                protein += line.strip('\n')
        all_proteins.append(protein)

    return all_ids, all_proteins


def write_search_proteome(proteome, output_folder, contams_db):
    """ Write new proteome file with decoys.

    Parameters
    ----------
    proteome : str
        Path to a proteome fasta file.
    output_folder : str
        Path to the inSPIRE output folder.
    contams_db : str
        Path to a contaminants fasta file to be appended to the search.
    """
    all_ids = []
    all_proteins = []

    all_ids, all_proteins = get_proteins(proteome, all_ids, all_proteins)
    if contams_db is not None:
        all_ids, all_proteins = get_proteins(contams_db, all_ids, all_proteins)

    # Copy original proteome file.
    shutil.copyfile(proteome, f'{output_folder}/search_proteome.fasta')
    if contams_db is not None:
        with open(
            f'{output_folder}/search_proteome.fasta',
            mode='a+',
            encoding='UTF-8',
        ) as search_file:
            with open(contams_db, mode='r', encoding='UTF-8') as cont_file:
                search_file.write(cont_file.read())

    # Append decoys to copy of original proteome.
    with open(
        f'{output_folder}/search_proteome.fasta',
        mode='a',
        encoding='UTF-8',
    ) as out_file:
        for prot_id, prot_seq in zip(all_ids, all_proteins):
            rev_sequence = prot_seq[::-1]
            out_file.write(
                f'>rev_{prot_id}\n{rev_sequence}\n'
            )


def write_fragger_params(config, fragger_params_template):
    """ Functiont to write MSFragger parameters.

    Parameters
    ----------
    config : inSPIRE.config.Config
        Config object for the experiment.
    fragger_params_template : str
        Path to the template for MSFragger parameters.
    """
    if config.mz_units == 'Da':
        ms2_units = 0
    else:
        ms2_units = 1

    if config.enzyme == 'trypsin':
        cleavage = TRYPTIC_CLEAVAGE
    else:
        cleavage = UNSPECIFIC_CLEAVAGE

    with open(
        fragger_params_template,
        mode='r',
        encoding='UTF-8',
    ) as frag_template_file:
        fragger_params = frag_template_file.read().format(
            search_database=f'{config.output_folder}/search_proteome.fasta',
            ncpus=config.n_cores,
            precursor_tolerance=config.ms1_accuracy,
            fragament_tolerance=config.mz_accuracy,
            fragment_units=ms2_units,
            top_n_candidates=10,
            cleavage_parameters=cleavage,
        )

    with open(
        f'{config.output_folder}/fragger.params',
        mode='w',
        encoding='UTF-8',
    ) as params_file:
        params_file.write(fragger_params)


def clean_up_fragger(config):
    """ Function to clean up files after MSFragger execution.

    Parameters
    ----------
    config : inSPIRE.config.Config
        Config object for the experiment.
    """
    with open(
        f'{config.output_folder}/fragger_searches.txt',
        mode='w',
        encoding='UTF-8',
    ) as frag_out:
        for s_file in os.listdir(config.scans_folder):
            if s_file.endswith('.pepXML'):
                frag_out.write(f'{config.scans_folder}/{s_file}\n')

    # Remove intermediate files
    mzml_files = [
        f'{config.scans_folder}/{scan_f}' for scan_f in os.listdir(config.scans_folder)
        if scan_f.lower().endswith('mzML')
    ]
    for mzml_file in mzml_files:
        os.remove(mzml_file)

def execute_msfragger(config):
    """ Function to run an MSFragger search with inSPIRE default settings.

    Parameters
    ----------
    config : inSPIRE.config.Config
        Config object for the experiment.
    """
    home = str(Path.home())
    fragger_params_template = config.fragger_params
    if config.fragger_use_contams:
        contams_db = (
            f'{home}/inSPIRE_models/utilities/contaminants_20120713.fasta'
        )
    else:
        contams_db = None
    frag_pipe_script_path = (
        f'{home}/inSPIRE_models/utilities/fragpipe_ms_fragger_script.py'
    )
    raw_files_for_fragger = ' '.join([
        f'{config.scans_folder}/{scan_f}' for scan_f in os.listdir(config.scans_folder)
        if scan_f.lower().endswith('.raw')
    ])

    write_search_proteome(config.proteome, config.output_folder, contams_db)
    write_fragger_params(config, fragger_params_template)

    os.system(
        f'{sys.executable} {frag_pipe_script_path} {config.fragger_db_splits} ' +
        f' "java -Xmx{config.fragger_memory}g -jar" ' +
        f' {config.fragger_path} {config.output_folder}/fragger.params ' +
        f' {raw_files_for_fragger} > {config.output_folder}/fragger.log '
    )

    clean_up_fragger(config)



""" Function for creating sequence logo plots.
"""
from math import ceil

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import fisher_exact
from statsmodels.sandbox.stats.multicomp import multipletests

from inspire.custom_logo import CustomLogo


AMINO_ACIDS = 'ACDEFGHIKLMNPQRSTVWY'
AA_COLOUR_SCHEME = {
    'P': 'deeppink',
    'M': 'orange',
    'A': 'orange',
    'V': 'orange',
    'I': 'orange',
    'L': 'orange',
    'F': 'orange',
    'Y': 'orange',
    'W': 'orange',
    'H': 'seagreen',
    'R': 'seagreen',
    'K': 'seagreen',
    'D': 'firebrick',
    'E': 'firebrick',
    'N': 'dodgerblue',
    'Q': 'dodgerblue',
    'S': 'dodgerblue',
    'T': 'dodgerblue',
    'G': 'dodgerblue',
    'C': 'dodgerblue',
}


def create_comparison_logo_plot(data_frame_list, title_list, peptide_length, output_folder, out_name, data_index=None):
    """ Function to compare frequencies of amino acids
    """
    dim_val = len(data_frame_list)-1
    n_plots = (dim_val ** 2 + dim_val)//2
    _, axes = plt.subplots(1, n_plots, figsize=(5*n_plots, 5))

    if n_plots == 1:
        axes = [axes]

    plot_idx = 0

    for col_idx, stata_df1 in enumerate(data_frame_list[:-1]):
        for row_idx, strata_df2 in enumerate(data_frame_list[col_idx+1:]):
            count_df_1 = get_count_df(stata_df1, peptide_length)
            count_df_2 = get_count_df(strata_df2, peptide_length)

            js_div, entropies = jensenshannon(
                count_df_1,
                count_df_2,
                axis=1,
            )

            entropy_df = pd.DataFrame(entropies)
            entropy_df.columns = list(AMINO_ACIDS)
            entropy_df.index += 1


            entropy_df['jsDivergence'] = js_div
            entropy_df = entropy_df.apply(scale_to_js_divergence, axis=1)
            entropy_df = entropy_df.drop(['jsDivergence'], axis=1)
            p_val_df = get_fishers_exact_pvals(count_df_1, count_df_2, peptide_length)
            signif_dict = get_significance_dict(p_val_df)

            logo_plot, axes[plot_idx] = create_logo_plot(
                entropy_df, axes[plot_idx], peptide_length, signif_dict
            )

            logo_plot.ax.set_title(f'{title_list[col_idx]} vs. {title_list[col_idx+row_idx+1]}')
            max_val = np.abs(js_div).max()
            if max_val < 0.2:
                cut_off = ceil(max_val*20)/20
                logo_plot.ax.set_ylim([-cut_off, cut_off])
                y_ticks = [x for x in [
                    -0.2, -0.15, -0.1, -0.05, 0, 0.05, 0.1, 0.15, 0.2
                ] if abs(x) <= cut_off]
                logo_plot.ax.set_yticks(y_ticks)
            else:
                logo_plot.ax.set_ylim([-ceil(max_val*5)/5, ceil(max_val*5)/5])
            
            if data_index is not None:
                logo_plot.ax.set_xlim([0.5, peptide_length+0.5])
                logo_plot.ax.set_xticklabels(data_index)
                logo_plot.ax.set_xlabel('Cluster Compared')
            plot_idx += 1

    axes[0].set_ylabel('JS Divergence')

    plt.tight_layout()
    plt.savefig(f'{output_folder}/{out_name}.svg', format='svg')


def get_count_df(strat_df, peptide_length):
    """ Function to get counts of amino acids at each position for peptide of a
        given length.
    """
    aa_counts = np.zeros([peptide_length, len(AMINO_ACIDS)])
    for _, df_row in strat_df.iterrows():
        for pos_idx, amino_acid in enumerate(df_row['peptide']):
            if amino_acid == 'X':
                continue
            aa_counts[pos_idx, AMINO_ACIDS.index(amino_acid)] += 1

    count_df = pd.DataFrame(
        aa_counts,
        columns=list(AMINO_ACIDS)
    )
    count_df.index += 1

    return count_df


def jensenshannon(p, q, base=None, *, axis=0, keepdims=False):
    """ Modified js divergence from scipy.
    """
    p = np.asarray(p)
    q = np.asarray(q)
    p = p / np.sum(p, axis=axis, keepdims=True)
    q = q / np.sum(q, axis=axis, keepdims=True)
    m = (p + q) / 2.0
    relative_entropy_fn = np.vectorize(_element_wise_rel_entropy)
    left = relative_entropy_fn(p, m)
    right = relative_entropy_fn(q, m)

    left_sum = np.sum(left, axis=axis, keepdims=keepdims)
    right_sum = np.sum(right, axis=axis, keepdims=keepdims)
    js = left_sum + right_sum
    if base is not None:
        js /= np.log(base)

    return (js / 2.0), left-right


def _element_wise_rel_entropy(x, y):
    """ Function to calculate entropy per element.
    """
    if x > 0 and y > 0:
        return x*np.log(x/y)
    if x == 0 and y >= 0:
        return 0.0
    return np.inf


def get_significance_dict(p_val_df):
    """ Helper function to get out the significantce into a dictionary.
    """
    sig_dict = {}
    for _, df_row in p_val_df.iterrows():
        sig_dict[f'{df_row["residue"]}{df_row["position"]}'] = df_row['significant']
    return sig_dict


def scale_to_js_divergence(df_row):
    """ Function to scale positive and negative entropies to JS divergence.
    """
    sum_neg = 0.0
    sum_pos = 0.0
    for a_a in AMINO_ACIDS:
        if df_row[a_a] > 0:
            sum_pos += df_row[a_a]
        elif df_row[a_a] < 0:
            sum_neg += abs(df_row[a_a])

    for a_a in AMINO_ACIDS:
        if df_row[a_a] > 0:
            df_row[a_a] = df_row['jsDivergence']*df_row[a_a]/sum_pos
        elif df_row[a_a] < 0:
            df_row[a_a] = df_row['jsDivergence']*df_row[a_a]/sum_neg
    return df_row


def create_logo_plot(amino_acid_data, axis, peptide_length, custom_alphas=None):
    """ Function to clean style on plots.
    """
    logo_plot = CustomLogo(
        amino_acid_data,
        ax=axis,
        font_name='DejaVu Sans',
        color_scheme=AA_COLOUR_SCHEME,
        flip_below=False,
        vpad=.2,
        width=.8,
        custom_alphas=custom_alphas,
    )
    logo_plot.style_xticks(anchor=1, spacing=1)
    logo_plot.ax.set_xlim([0, peptide_length+1])

    # Hide the right and top spines
    logo_plot.ax.spines.right.set_visible(False)
    logo_plot.ax.spines.top.set_visible(False)

    # Only show ticks on the left and bottom spines
    logo_plot.ax.yaxis.set_ticks_position('left')
    logo_plot.ax.xaxis.set_ticks_position('bottom')

    return logo_plot, axis

def get_fishers_exact_pvals(count_df_1, count_df_2, peptide_length):
    """ Function to get fishers exact p values.
    """
    p_val_list = []

    sums_1 = count_df_1.sum(axis=1)
    sums_2 = count_df_2.sum(axis=1)

    for a_a in AMINO_ACIDS:
        for idx in range(1,peptide_length+1):
            group_1_aa = count_df_1[a_a].iloc[idx-1]
            group_2_aa = count_df_2[a_a].iloc[idx-1]

            group_1_notaa = sums_1.iloc[idx-1] - group_1_aa
            group_2_notaa = sums_2.iloc[idx-1] - group_2_aa

            fishers_array = np.array([
                [group_1_aa, group_2_aa],
                [group_1_notaa, group_2_notaa],
            ])

            _, pvalue = fisher_exact(fishers_array)
            p_val_list.append({
                'position': idx,
                'residue': a_a,
                'count1': group_1_aa,
                'count2': group_2_aa,
                'count1_false': group_1_notaa,
                'count2_false': group_2_notaa,
                'pvalue': pvalue,
            })

    pvalue_df = pd.DataFrame(p_val_list)
    pvalue_df['adjusted_pValue'] = multipletests(pvalue_df['pvalue'], method='fdr_bh')[1]
    pvalue_df['significant'] = pvalue_df['pvalue'].apply(lambda x : 1 if x < 0.05 else 0.5)

    return pvalue_df



""" Functions for rescoring PSMs with an optimised feature set.
"""
from pathlib import Path
import subprocess

import pandas as pd
import polars as pl

from inspire.constants import (
    ACCESSION_STRATUM_KEY,
    ACCESSION_KEY,
    CHARGE_KEY,
    ENDC_TEXT,
    ENGINE_SCORE_KEY,
    FINAL_POSTEP_KEY,
    FINAL_Q_VALUE_KEY,
    FINAL_SCORE_KEY,
    OKCYAN_TEXT,
    OUT_ACCESSION_KEY,
    OUT_POSTEP_KEY,
    OUT_Q_KEY,
    OUT_SCORE_KEY,
    PEPTIDE_KEY,
    PSM_ID_KEY,
    RT_KEY,
    SCAN_KEY,
    SOURCE_KEY,
    SPECTRAL_ANGLE_KEY,
)
from inspire.input.mhcpan import read_mhcpan_output
from inspire.utils import fetch_proteome, parallel_remap

def apply_rescoring(
        output_folder,
        input_filename,
        fdr,
        rescore_method,
        output_prefix,
        rescore_command,
        proteome=None,
        decoy_prot_key='rev_',
    ):
    """ Function to apply percolator and return the PSMs matched.

    Parameters
    ----------
    perc_input_df : pd.DataFrame
        A DataFrame read for percolator input.
    output_folder : str
        The folder in which all output for the pipeline is written.
    fdr : float
        The False Positive Rate with which to train percolator (0.0-1.0)
    output_filename : str or None
        A specific filename for the output psms, defaults
    output_weights : bool
        Flag indicating whether or not to output the feature weights.

    Returns
    -------
    results : pd.DataFrame
        The predictions from Percolator.
    """
    psm_output_key = f'{output_folder}/{output_prefix}.{rescore_method}.psms.txt'
    pep_output_key = f'{output_folder}/{output_prefix}.{rescore_method}.peptides.txt'
    prot_out_key = f'{output_folder}/{output_prefix}.{rescore_method}.proteins.txt'

    if rescore_method == 'mokapot':
        clis = (
            f' --dest_dir {output_folder} --keep_decoys  ' +
            f' --train_fdr {fdr} ' +
            f' --test_fdr {fdr} --file_root {output_prefix} --save_models '
        )
        trailing_args = ''
    elif rescore_method == 'percolatorSeparate':
        percolator_decoy_key = f'{output_folder}/{output_prefix}.{rescore_method}.decoy.psms.txt'
        weights_path = f'{output_folder}/{output_prefix}.{rescore_method}.weights.csv'
        clis = (
            f' -F {fdr} -t {fdr} -i 10 -M {percolator_decoy_key} --post-processing-tdc  ' +
            f' -w {weights_path} ' +
            f' --results-psms {psm_output_key} --results-peptides {pep_output_key} '
        )
        trailing_args = ''
    else:
        percolator_decoy_key = f'{output_folder}/{output_prefix}.{rescore_method}.decoy.psms.txt'
        weights_path = f'{output_folder}/{output_prefix}.{rescore_method}.weights.csv'
        clis = (
            f' -F {fdr} -t {fdr} -i 10 -M {percolator_decoy_key} --post-processing-tdc ' +
            f' -I concatenated -w {weights_path} ' +
            f' --results-psms {psm_output_key} --results-peptides {pep_output_key} '
        )
        trailing_args = ''

    if proteome is not None:
        trailing_args += (
            f' -f {proteome} -z no_enzyme --results-proteins {prot_out_key} -P {decoy_prot_key}'
        )

    bash_command = (
        f'"{rescore_command}" {clis} {output_folder}/{input_filename} {trailing_args}'
    )

    with open(f'{output_folder}/rescore.log', 'w', encoding='UTF-8') as log_file:
        subprocess.run(
            bash_command,
            check=True,
            shell=True,
            stdout=log_file,
        )

    psm_results_df = pl.read_csv(psm_output_key, separator='\t')
    psm_results_df = psm_results_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(lambda x : x[2:-2])
    )

    peptide_results_df = pl.read_csv(pep_output_key, separator='\t')
    peptide_results_df = peptide_results_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(lambda x : x[2:-2])
    )

    return psm_results_df, peptide_results_df

def _split_psm_ids(psm_id):
    """ Function for splitting a PSM Id back into its source name, scan number
        and peptide sequence.

    Parameters
    ----------
    df_row : pd.Series
        A row of the DataFrame to which the function is being applied.

    Parameters
    ----------
    df_row : pd.Series
        The same row with source and scan added.
    """
    results = {}
    source_scan_list = psm_id.split('_')
    results['modifiedSequence'] = source_scan_list[-1]
    results[SCAN_KEY] = int(source_scan_list[-2])
    results[SOURCE_KEY] = '_'.join(source_scan_list[:-2])
    return results

def _regroup_accession(df_row, acc_cols):
    """ Helper function to remove one hot encoding from Accession Stratum.

    Parameters
    ----------
    df_row : pd.Series
        A row of the final results DataFrame.
    acc_cols : list of str
        All of the accession related columns.

    Returns
    -------
    acc_stratum : str
        The final accession stratum listed in the results.
    """
    for acc_col in acc_cols:
        if df_row[acc_col] == 1:
            return acc_col.split('_')[1]
    return 'unknown'

def _add_key_features(target_psms, config):
    """ Function to add spectral angle and engine score back to percolator
        output PSMs.

    Parameters
    ----------
    target_psms : pd.DataFrame
        A DataFrame of percolator output PSMs.
    config : inspire.config.Config
        The config object used throughout the pipeline.

    Returns
    -------
    output_df : pd.DataFrame
        The output PSMs labelled with original search engine score and
        spectral angle.
    """
    psm_id_key = PSM_ID_KEY[config.rescore_method]

    input_key = f'{config.output_folder}/input_all_features.tab'
    input_df = pl.from_pandas(pd.read_csv(
        input_key,
        sep='\t',
    ))
    input_df = input_df.unique(subset=[psm_id_key, PEPTIDE_KEY])

    key_features = [
        SPECTRAL_ANGLE_KEY,
        'spearmanR',
        'predCoverage',
        'matchedCoverage',
        RT_KEY,
        'deltaRT',
        ENGINE_SCORE_KEY,
        CHARGE_KEY,
    ]
    if config.use_accession_stratum:
        acc_cols = [
            x for x in input_df.columns if x.startswith('accession') and x != 'accessionGroup'
        ]
        input_df = input_df.with_columns(
            pl.struct(acc_cols).apply(
                lambda x : _regroup_accession(x, acc_cols)
            ).alias(ACCESSION_STRATUM_KEY)
        )

        key_features.append(ACCESSION_STRATUM_KEY)

    if config.use_binding_affinity is not None:
        mhc_pan_df = read_mhcpan_output(
            f'{config.output_folder}/mhcpan',
            per_allele=True,
        )
        mhc_pan_df = mhc_pan_df.rename({
            'Peptide': PEPTIDE_KEY,
        })
        mhc_pan_cols = (
            [x for x in mhc_pan_df.columns if (
                x.endswith('Affinity') or x.endswith('Level') or x.endswith('%Rank_BA')
            )]
        )

        target_psms = target_psms.join(
            mhc_pan_df.select([PEPTIDE_KEY] + mhc_pan_cols),
            on=PEPTIDE_KEY,
            how='left',
        )

    if isinstance(config.collision_energy, list):
        key_features.append('collisionEnergy')

    output_df = target_psms.join(
        input_df[[psm_id_key, PEPTIDE_KEY] + key_features],
        how='inner',
        on=[psm_id_key, PEPTIDE_KEY]
    )

    if config.use_binding_affinity is not None:
        for col in mhc_pan_cols:
            if col.endswith('Affinity') or col.endswith('%Rank_BA'):
                output_df = output_df.with_columns(
                    pl.col(col).fill_null(
                        pl.lit(-1),
                    ).alias(col),
                )
            else:
                output_df = output_df.with_columns(
                    pl.col(col).apply(
                        lambda x : 'Strong-Binder' if x == '<=SB' else (
                            'Weak-Binder' if x == '<=WB' else (
                                'Not predicted' if x is None else
                            'Non-Binder')
                        ),
                        skip_nulls=False,
                    )
                )
        key_features.extend(mhc_pan_cols)

    return output_df, key_features

def final_rescoring(config):
    """ Function to rescore PSMs using the final feature set.

    Parameters
    ----------
    config : inspire.config.Config
        The config object used throughout the pipeline.
    """
    in_path = 'final_input.tab'
    output_prefix = 'final'

    if config.infer_proteins:
        proteome = config.proteome
    else:
        proteome = None

    target_psms, target_peptides = apply_rescoring(
        config.output_folder,
        in_path,
        config.fdr,
        config.rescore_method,
        output_prefix,
        config.rescore_command,
        proteome,
        decoy_prot_key=config.decoy_protein_flag,
    )

    print(
        OKCYAN_TEXT + '\tRescoring complete.' + ENDC_TEXT
    )

    output_peptides_df = apply_post_processing(target_peptides, config)
    output_psm_df = apply_post_processing(target_psms, config)

    if config.accession_format == 'invitroSPI':
        output_peptides_df.write_csv(f'{config.output_folder}/pre_finalAssignments.csv')
    else:
        output_peptides_df.write_csv(f'{config.output_folder}/finalPeptideAssignments.csv')

    output_psm_df.write_csv(f'{config.output_folder}/finalPsmAssignments.csv')

    print(
        OKCYAN_TEXT + '\tFinal assignments written to csv.' + ENDC_TEXT
    )


def apply_post_processing(target_psms, config):
    """ Function to separate columns and re-add relevant information.
    """
    out_score_key = OUT_SCORE_KEY[config.rescore_method]
    out_q_key = OUT_Q_KEY[config.rescore_method]
    out_postep_key = OUT_POSTEP_KEY[config.rescore_method]
    psm_id_key = PSM_ID_KEY[config.rescore_method]
    out_accession_key = OUT_ACCESSION_KEY[config.rescore_method]

    if 'PSMId' in target_psms.columns:
        target_psms = target_psms.rename({'PSMId': psm_id_key})


    output_df, key_features = _add_key_features(target_psms, config)

    output_df = output_df.with_columns(
        pl.col(psm_id_key)
        .apply(_split_psm_ids, skip_nulls=False)
        .alias("results")
    ).unnest("results")

    output_df = output_df.rename({
        out_score_key: FINAL_SCORE_KEY,
        out_q_key: FINAL_Q_VALUE_KEY,
        out_postep_key: FINAL_POSTEP_KEY,
        out_accession_key: ACCESSION_KEY,
        out_postep_key: FINAL_POSTEP_KEY,
    })

    final_columns = (
        [
            SOURCE_KEY,
            SCAN_KEY,
            PEPTIDE_KEY,
            'modifiedSequence',
            FINAL_SCORE_KEY,
            FINAL_Q_VALUE_KEY,
            FINAL_POSTEP_KEY,
        ] + key_features
    )

    if config.map_contaminants is not None:
        if config.map_contaminants == 'standard':
            home = str(Path.home())
            contamns_path =f'{home}/inSPIRE_models/utilities/contaminants_20120713.fasta'
        else:
            contamns_path = config.map_contaminants


        if config.proteome is not None:
            target_proteome = fetch_proteome(config.proteome, with_desc=False)
            output_df = parallel_remap(
                output_df,
                config.n_cores,
                target_proteome,
                'mapsToTarget',
                trace_accession=False,
            )

            contams_proteome = fetch_proteome(contamns_path, with_desc=False)
            output_df = parallel_remap(
                output_df,
                config.n_cores,
                contams_proteome,
                'mapsToContaminant',
                trace_accession=False,
            )
        if config.proteome is not None:
            final_columns.extend(['mapsToTarget', 'mapsToContaminant'])


    final_columns.append(ACCESSION_KEY)
    output_df = output_df.select(final_columns)
    output_df = output_df.sort(by=[FINAL_SCORE_KEY, SOURCE_KEY, SCAN_KEY], descending=True)

    return output_df



""" Functions for calculating spectral features.
"""

# from copyreg import pickle
from copy import deepcopy
import gc
from math import acos, pi
from statistics import mean, variance

import polars as pl
import numpy as np
from scipy.stats import spearmanr, pearsonr
from sklearn.metrics import median_absolute_error

from inspire.constants import (
    CHARGE_KEY,
    INTENSITIES_KEY,
    LOSS_IONS_KEY,
    MAE_KEY,
    NEUTRAL_LOSSES,
    FRAG_MZ_ERR_MED_KEY,
    FRAG_MZ_ERR_VAR_KEY,
    MATCHED_IONS_KEY,
    NOT_ASSIGNED_KEY,
    PRECURSOR_INTE_KEY,
    MZS_KEY,
    PEARSON_KEY,
    PEPTIDE_KEY,
    PROSIT_IONS_KEY,
    PROSIT_INTES_KEY,
    PROSIT_SEQ_KEY,
    PROTON,
    PTM_ID_KEY,
    PTM_SEQ_KEY,
    PTM_WEIGHT_KEY,
    SPEARMAN_KEY,
    SPECTRAL_ANGLE_KEY,
)
from inspire.mz_match import get_ion_masses, match_mz
from inspire.prosit_delta import get_deltas
from inspire.utils import get_ox_flag


SPECTRAL_FEATURES = [
    SPECTRAL_ANGLE_KEY,
    SPEARMAN_KEY,
    PEARSON_KEY,
    MAE_KEY,
    FRAG_MZ_ERR_MED_KEY,
    FRAG_MZ_ERR_VAR_KEY,
    NOT_ASSIGNED_KEY,
    LOSS_IONS_KEY,
    PRECURSOR_INTE_KEY,
    'yIsDominantIonSeries',
    'bIsDominantIonSeries',
    'maxTypeSpectralAngle',
    'minTypeSpectralAngle',
    'fracMatchedKR',
    'possibleKrFragsDivTotal',
    'matchedCoverage',
    'minMatchedCoverage',
    'maxMatchedCoverage',
    'predCoverage',
    'predNotFoundCoverage',
    'nMajorMatchedDivFrags',
    'nMajorPredNotFoundDivFrags',
    'nMajorNotMatchableDivFrags',
    'spearmanMajorIons',
    'nMinorMatchedDivFrags',
    'nMinorNotMatchableDivFrags',
    'spearmanMinorIons',
    'spectrumDensity',
]

DELTA_FEATURES = [
    'prositDeltaQuartile1',
    'prositDeltaQuartile3',
    'prositDeltaMedian',
    'nDeltasAboveThreshold',
    'nDeltasAboveThresholdA',
    'nDeltasAboveZero',
    'maxPrositDelta',
    'minPrositDelta',
]

PROSIT_MAJOR_MINOR_CUT_OFF = 0.1
SPECTRUM_MAJOR_MINOR_CUT_OFF = 0.1

def calculate_spectral_angle(true, predicted):
    """ Function to calculate the spectral angle between the true and predicted
        spectra.

    Parameters
    ----------
    true : dict
        A dictionary of ion intensities from the true spectrum.
    predicted : dict
        A dictionary of ion intensities from the predicted spectrum.

    Returns
    -------
    spectral_angle : float
        The spectral angle between the two spectra.
    """
    true_l2_norm = np.linalg.norm(true, ord=2)
    pred_l2_norm = np.linalg.norm(predicted, ord=2)
    if true_l2_norm == 0 or pred_l2_norm == 0:
        return 0.0

    product = np.dot(true/true_l2_norm, predicted/pred_l2_norm)

    spectral_distance = 2*acos(product)/pi

    return 1.0 - spectral_distance

def check_for_precursor_peak(observed_mzs, obs_intes, precursor_mz, mz_accuracy, assigned_inds):
    """ Function to check for the presence of the precursor fragment in spectrum.

    Parameters
    ----------
    observed_mzs : np.array of float
        An array of mz values of fragments observed.
    precursor_mz : float
        The m/z of the precursor.
    mz_accuracy : float
        The m/z accuracy of the mass spectrometer.
    assigned_inds : list
        The indices already matched to an ion.

    Returns
    -------
    precursor_present : int
        Flag indicating if the precursor m/z is present.
    assigned_inds : list
        The input indices updated with the index of precursor matched ion.
    """
    matched_mz_ind = np.argmin(
        np.abs(observed_mzs - precursor_mz)
    )

    if abs(observed_mzs[matched_mz_ind] - precursor_mz) < mz_accuracy:
        assigned_inds.append(matched_mz_ind)
        return obs_intes[matched_mz_ind], assigned_inds
    return 0, assigned_inds


def get_matches(
        all_prosit_masses,
        prosit_intensities,
        observed_mzs,
        precursor_weight,
        observed_intensities,
        mz_accuracy,
        precursor_z,
        mz_units,
    ):
    """ Function to find all the matched mz measuremnets between the
        predicted MS2 spectrum of a peptide and an observed MS2 spectra.

    Parameters
    ----------
    expected_mzs : np.array
        An array of the mz of expected fragments that could be generated
        for the peptide.
    prosit_intensities : np.array
        The predicted intensities from prosit.
    observed_mzs : np.array
        An array of the mz of the observed fragments from the MS spectrum.
    observed_intensities : np.array
        An array of the intensities observed for the corresponding observed
        mzs.
    mz_accuracy : float
        The accuracy of the m/z measurement for the observations.

    Returns
    -------
    matches : np.array
        A 2d array of shape (len(observed_mzs), n ion types considered).
        If there is no match between the observed mz and a given ion type
        the value of the corresponding array entry is 0. Otherwise it will
        be the fragment index.
    """
    # Loop over observed fragments matching them to all possible prosit ions.
    ordered_prosit = np.array(list(prosit_intensities.keys()), dtype='object')
    final_intensities = np.zeros(len(prosit_intensities))
    observed_mzs = np.array(observed_mzs)
    observed_intensities = np.array(observed_intensities)

    possible_loss_ions = []
    mz_errors = []
    assigned_inds = []

    max_frag_charge = min(4, precursor_z+1)
    match_idx = 0
    for ion_type in all_prosit_masses:
        for charge in range(1, max_frag_charge):
            for fragment_idx in range(len(all_prosit_masses[ion_type])):
                ion_code = f'{ion_type}{fragment_idx+1}'
                if charge > 1:
                    ion_code += f'^{charge}'

                if ion_code not in ordered_prosit:
                    continue

                mass = all_prosit_masses[ion_type][fragment_idx] + (charge * PROTON)
                if mz_units == 'ppm':
                    mz_err = (mass/charge)*mz_accuracy*(10**-6)
                else:
                    mz_err = mz_accuracy

                mz_diff, matched_mz_ind = match_mz(
                    all_prosit_masses[ion_type][fragment_idx],
                    charge,
                    observed_mzs,
                    loss=0.0
                )

                if abs(mz_diff) < mz_err:
                    if ion_code in ordered_prosit:
                        match_idx = np.where(ordered_prosit == ion_code)[0]
                        final_intensities[match_idx] = observed_intensities[matched_mz_ind]
                        mz_errors.append(mz_diff)
                        for loss in NEUTRAL_LOSSES:
                            loss_mz_diff, loss_matched_mz_ind = match_mz(
                                all_prosit_masses[ion_type][fragment_idx],
                                precursor_z,
                                observed_mzs,
                                loss=loss
                            )
                            if abs(loss_mz_diff) < mz_err:
                                possible_loss_ions.append(
                                    observed_intensities[loss_matched_mz_ind]
                                )
                        assigned_inds.append(matched_mz_ind)

    precursor_mz = (precursor_weight + (PROTON*precursor_z))/precursor_z

    if mz_units == 'ppm':
        mz_err = precursor_mz*mz_accuracy*(10**-6)
    else:
        mz_err = mz_accuracy

    precursor_inte, assigned_inds = check_for_precursor_peak(
        observed_mzs,
        observed_intensities,
        precursor_mz,
        mz_err,
        assigned_inds,
    )

    # Select the intensities not assigned to anything.
    unassigned_inds = [
        x for x in range(len(observed_intensities)) if x not in assigned_inds
    ]
    unassigned_intes = observed_intensities[unassigned_inds]

    match_info = {
        'matched_intensities': final_intensities,
        'ordered_prosit_ions': ordered_prosit,
        'ordered_prosit_intes': np.array(
            [prosit_intensities[x] for x in ordered_prosit]
        ),
    }

    return (
        match_info,
        precursor_inte,
        unassigned_intes,
        mz_errors,
        possible_loss_ions,
    )

def get_mz_error_stats(mz_errors, mz_accu):
    """ Function to get the median and variance of the mz errors on prosit matched fragments.

    Parameters
    ----------
    mz_errors : list of float
        The mz error for each matched ion.

    Returns
    -------
    median_mz_error : float
        The median mz error on the matched ions.
    mz_error_variance : float
        The variance of the mz errors.
    """
    # Get the error in matching each fragment m/z.
    if mz_errors:
        if len(mz_errors) > 1:
            mz_error_variance = variance(mz_errors)
        else:
            mz_error_variance = mz_accu
        abs_mz_errors = [abs(x) for x in mz_errors]
        median_mz_error = mean(abs_mz_errors)
        return median_mz_error, mz_error_variance
    return mz_accu, mz_accu

def get_coverage(pep_len, ion_array, per_letter=False):
    """ Function to get the coverage for an ion series.
    """
    y_inds = [
        int(ion.split('^')[0][1:]) for ion in ion_array if ion.startswith('y')
    ]
    b_inds = [
        int(ion.split('^')[0][1:]) for ion in ion_array if ion.startswith('b')
    ]
    y_rev_inds = [pep_len-x for x in y_inds]
    total_coverage = len(set(b_inds+y_rev_inds))/(pep_len-1)

    if per_letter:
        b_coverage = len(set(b_inds))/(pep_len-1)
        y_coverage = len(set(y_inds))/(pep_len-1)
        return total_coverage, b_coverage, y_coverage
    return total_coverage

def calculate_spectral_features(
        df_row,
        ptm_id_weights,
        mz_accuracy,
        mz_units,
        model,
        ox_flag,
        delta_method,
        minimal_features=False,
    ):
    """ Function to extract the ion intensities from the true spectra which match
    """
    results = {}
    sequence = df_row[PEPTIDE_KEY]
    seq_len = len(sequence)
    n_frags_possible = seq_len - 1
    mz_array = np.array(df_row[MZS_KEY])
    intes_array = np.array(df_row[INTENSITIES_KEY])

    prosit_preds = dict(zip(df_row[PROSIT_IONS_KEY], df_row[PROSIT_INTES_KEY]))

    potential_ion_mzs, precursor_weight = get_ion_masses(
        sequence,
        ptm_id_weights,
        df_row[PTM_SEQ_KEY]
    )

    (
        match_info, precursor_inte, unassigned_intensities, mz_errors, possible_alts
    ) = get_matches(
        potential_ion_mzs,
        prosit_preds,
        mz_array,
        precursor_weight,
        df_row[INTENSITIES_KEY],
        mz_accuracy,
        df_row[CHARGE_KEY],
        mz_units,
    )
    matched_intensities = match_info['matched_intensities']
    ordered_prosit_ions = match_info['ordered_prosit_ions']
    ordered_prosit_intes = match_info['ordered_prosit_intes']
    truly_matched_intes = matched_intensities[matched_intensities > 0.0]

    median_mz_error, mz_error_variance = get_mz_error_stats(mz_errors, min(mz_accuracy, 0.04))

    matched_l2_norm = np.linalg.norm(matched_intensities, ord=2)
    total_l2_norm = np.linalg.norm(df_row[INTENSITIES_KEY], ord=2)

    if matched_l2_norm:
        normed_matched_intensities = matched_intensities/matched_l2_norm
    else:
        normed_matched_intensities = matched_intensities

    ordered_matched_ions = ordered_prosit_ions[matched_intensities > 0.0]

    results[SPECTRAL_ANGLE_KEY] = calculate_spectral_angle(
        normed_matched_intensities,
        ordered_prosit_intes,
    )

    results = get_coverage_features(
        results,
        seq_len,
        ordered_matched_ions,
        ordered_prosit_ions,
    )

    if len(truly_matched_intes) > 0:
        spearman_total = spearmanr(normed_matched_intensities, ordered_prosit_intes)[0]
        pearson_total = pearsonr(normed_matched_intensities, ordered_prosit_intes)[0]
        if not np.isnan(spearman_total):
            results[SPEARMAN_KEY] = float(spearman_total)
        else:
            results[SPEARMAN_KEY] = 0.0
        if not np.isnan(pearson_total):
            results[PEARSON_KEY] = float(pearson_total)
        else:
            results[PEARSON_KEY] = -1.0
    else:
        results[SPEARMAN_KEY] = 0.0
        results[PEARSON_KEY] = 0.0

    if minimal_features:
        return results

    mz_range = (mz_array.max() - mz_array.min())
    if mz_range > 0:
        results['spectrumDensity'] = len(mz_array)/mz_range
    else:
        results['spectrumDensity'] = 1.0

    major_pred_inds = ordered_prosit_intes >= PROSIT_MAJOR_MINOR_CUT_OFF

    major_prosit_preds = ordered_prosit_intes[major_pred_inds]
    minor_prosit_preds = ordered_prosit_intes[~major_pred_inds]

    major_matched_ions = normed_matched_intensities[major_pred_inds]
    minor_matched_ions = normed_matched_intensities[~major_pred_inds]

    n_major_pred = len(major_prosit_preds)

    if matched_l2_norm:
        major_prosit_l2_norm = np.linalg.norm(major_prosit_preds, ord=2)
        minor_prosit_l2_norm = np.linalg.norm(minor_prosit_preds, ord=2)

        major_l2_norm = np.linalg.norm(major_matched_ions, ord=2)
        minor_l2_norm = np.linalg.norm(minor_matched_ions, ord=2)

        n_major_matched = len(major_matched_ions[major_matched_ions > 0.0])
        n_minor_matched = len(truly_matched_intes) - n_major_matched

        if major_l2_norm:
            major_matched_ions /= major_l2_norm
        if minor_l2_norm:
            minor_matched_ions /= minor_l2_norm
        if major_prosit_l2_norm:
            major_prosit_preds /= major_prosit_l2_norm
        if minor_prosit_l2_norm:
            minor_prosit_preds /= minor_prosit_l2_norm

        if n_major_matched > 0:
            spearman_major = spearmanr(major_matched_ions, major_prosit_preds)[0]
            if not np.isnan(spearman_major):
                results['spearmanMajorIons'] = float(spearman_major)
            else:
                results['spearmanMajorIons'] = 0.0
        else:
            results['spearmanMajorIons'] = 0.0

        if n_minor_matched > 0:
            spearman_minor = spearmanr(minor_matched_ions, minor_prosit_preds)[0]
            if not np.isnan(spearman_minor):
                results['spearmanMinorIons'] = float(spearman_minor)
            else:
                results['spearmanMinorIons'] = 0.0
        else:
            results['spearmanMinorIons'] = 0.0

    else:
        n_major_matched = 0.0
        n_minor_matched = 0.0
        results['spearmanMajorIons'] = 0.0
        results['spearmanMinorIons'] = 0.0

    n_major_not_matchable = len(
        [x for x in unassigned_intensities if x/total_l2_norm > SPECTRUM_MAJOR_MINOR_CUT_OFF]
    )
    n_minor_not_matchable = len(unassigned_intensities)


    results['nMajorPredNotFoundDivFrags'] = (n_major_pred - n_major_matched)/n_frags_possible

    results[MAE_KEY] = float(median_absolute_error(
        normed_matched_intensities,
        ordered_prosit_intes,
    ))

    results['nMajorMatchedDivFrags'] = n_major_matched/n_frags_possible
    results['nMinorMatchedDivFrags'] = n_minor_matched/n_frags_possible
    results[MATCHED_IONS_KEY] = len(truly_matched_intes)/n_frags_possible

    results['nMajorNotMatchableDivFrags'] = n_major_not_matchable/n_frags_possible
    results['nMinorNotMatchableDivFrags'] = n_minor_not_matchable/n_frags_possible

    results[NOT_ASSIGNED_KEY] = len(unassigned_intensities)/n_frags_possible
    results[LOSS_IONS_KEY] = len(possible_alts)/n_frags_possible


    if precursor_inte:
        results[PRECURSOR_INTE_KEY] = float(precursor_inte/(matched_l2_norm+precursor_inte))
    else:
        results[PRECURSOR_INTE_KEY] = 0.0

    results[FRAG_MZ_ERR_MED_KEY] = float(median_mz_error)
    results[FRAG_MZ_ERR_VAR_KEY] = float(mz_error_variance)

    results = get_kr_feats(
        sequence, matched_intensities, ordered_prosit_intes, ordered_prosit_ions, results
    )

    pred_not_found_ions = ordered_prosit_ions[matched_intensities == 0.0]

    results['predNotFoundCoverage'] = get_coverage(
        seq_len, pred_not_found_ions
    )

    y_filter = [
        idx for idx in range(len(ordered_prosit_ions)) if ordered_prosit_ions[idx][0] == 'y'
    ]
    b_filter = [
        idx for idx in range(len(ordered_prosit_ions)) if ordered_prosit_ions[idx][0] == 'b'
    ]
    if not truly_matched_intes.size:
        results['maxTypeSpectralAngle'] = 0.0
        results['minTypeSpectralAngle'] = 0.0
    else:
        y_spectral_angle = calculate_spectral_angle(
            normed_matched_intensities[y_filter],
            ordered_prosit_intes[y_filter],
        )
        b_spectral_angle = calculate_spectral_angle(
            normed_matched_intensities[b_filter],
            ordered_prosit_intes[b_filter],
        )

        if results['bIsDominantIonSeries'] == 1:
            results['maxTypeSpectralAngle'] = b_spectral_angle
            results['minTypeSpectralAngle'] = y_spectral_angle
        elif results['yIsDominantIonSeries'] == 1:
            results['maxTypeSpectralAngle'] = y_spectral_angle
            results['minTypeSpectralAngle'] = b_spectral_angle
        else:
            results['maxTypeSpectralAngle'] = max([b_spectral_angle, y_spectral_angle])
            results['minTypeSpectralAngle'] = min([b_spectral_angle, y_spectral_angle])

    precursor_mz = (precursor_weight + (PROTON*df_row[CHARGE_KEY]))/df_row[CHARGE_KEY]
    if mz_units == 'ppm':
        mz_err = precursor_mz*mz_accuracy*(10**-6)
    else:
        mz_err = mz_accuracy

    if delta_method == 'predictor':
        matched_dict = dict(zip(ordered_matched_ions.tolist(), truly_matched_intes.tolist()))

        results = get_deltas(
            results,
            df_row,
            mz_array,
            intes_array,
            prosit_preds,
            ordered_prosit_ions,
            normed_matched_intensities,
            ordered_prosit_intes,
            model,
            ox_flag,
            mz_err,
            matched_dict,
        )

    return results

def get_coverage_features(
        results,
        seq_len,
        ordered_matched_ions,
        ordered_prosit_ions
    ):
    """ Function to get coverage features.
    """
    results['predCoverage'], b_pred_coverage, y_pred_coverage = get_coverage(
        seq_len, ordered_prosit_ions, per_letter=True
    )

    results['matchedCoverage'], b_matched_coverage, y_matched_coverage = get_coverage(
        seq_len, ordered_matched_ions, per_letter=True
    )

    if b_pred_coverage < y_pred_coverage:
        results['minMatchedCoverage'] = b_matched_coverage
        results['maxMatchedCoverage'] = y_matched_coverage
        results['yIsDominantIonSeries'] = 1
        results['bIsDominantIonSeries'] = 0
    elif b_pred_coverage > y_pred_coverage:
        results['yIsDominantIonSeries'] = 0
        results['bIsDominantIonSeries'] = 1
        results['minMatchedCoverage'] = y_matched_coverage
        results['maxMatchedCoverage'] = b_matched_coverage
    else:
        results['minMatchedCoverage'] = min([y_matched_coverage, b_matched_coverage])
        results['maxMatchedCoverage'] = max([y_matched_coverage, b_matched_coverage])
        results['yIsDominantIonSeries'] = 0
        results['bIsDominantIonSeries'] = 0

    return results


def get_kr_feats(sequence, matched_intensities, ordered_prosit_intes, ordered_prosit_ions, results):
    """ Function to get the difference between the spectral angle of.
    """
    seq_len = len(sequence)
    y_kr = seq_len
    b_kr = seq_len
    n_possible_kr = 0
    for idx, entry in enumerate(sequence):
        if entry in 'KR':
            b_kr = seq_len - (idx+2)
            n_possible_kr = b_kr
            break

    if b_kr < seq_len:
        for idx, entry in enumerate(sequence[::-1]):
            if entry in 'KR':
                y_kr = seq_len - (idx+2)
                n_possible_kr += y_kr
                break

    kr_inds = []
    if b_kr < seq_len:
        for idx, ion_name in enumerate(ordered_prosit_ions):
            ion_code = ion_name[0]
            frag_idx = int(ion_name[1:].split('^')[0])
            if ion_code == 'b':
                if frag_idx > b_kr:
                    kr_inds.append(idx)
            else:
                if frag_idx > y_kr:
                    kr_inds.append(idx)

        kr_prosit_intensities = ordered_prosit_intes[kr_inds]
        kr_matched_intensities = matched_intensities[kr_inds]

        if kr_prosit_intensities.size:
            results['fracMatchedKR'] = len(
                kr_matched_intensities[kr_matched_intensities > 0]
            )/len(kr_prosit_intensities)
        else:
            results['fracMatchedKR'] = len(
                matched_intensities[matched_intensities > 0]
            )/len(ordered_prosit_ions)
        results['possibleKrFragsDivTotal'] = n_possible_kr/(seq_len-1)

        return results

    results['fracMatchedKR'] = 0.0
    results['possibleKrFragsDivTotal'] = 0.0

    return results

def fetch_mod_weight_dict(mods_df):
    """ Function to fetch a dictionary of the weights of all modifications.

    Parameters
    ----------
    mods_df : pd.DataFrame
        A DataFrame of the modifications present in data.

    Returns
    -------
    ptm_id_weights : dict
        A dict mapping ptm ids to the change in mass cause by it.
    """
    if mods_df.empty:
        return {0: 0.0}
    mod_ids = mods_df[PTM_ID_KEY]
    mod_weights = mods_df[PTM_WEIGHT_KEY]

    ptm_id_weights = dict(zip(mod_ids, mod_weights))
    ptm_id_weights[0] = 0.0 # Add entry for no modification.

    return ptm_id_weights


def create_spectral_features(mods_df, config, task_id, model):
    """ Function to calculate spectral features between experimental and prosit predicted
        spectra.

    Parameters
    ----------
    spectral_df : pd.DataFrame
        A DataFrame containing experimental and prosit predicted spectra.
    mods_df : pd.DataFrame
        A small DataFrame detailing the unique modifications seen in the data.
    mz_accuracy : float
        The m/z accuracy of the instrument which measured the spectra.

    Returns
    -------
    spectral_df : pd.DataFrame
        The input DataFrame with features added to describe the match between experimental
        and prosit predicted spectra.
    """
    spectral_df = pl.read_parquet(
        f'{config.output_folder}/temp_{task_id}_in.parquet'
    )

    ox_flag = get_ox_flag(mods_df)

    ptm_id_weights = fetch_mod_weight_dict(mods_df)

    spectral_df = spectral_df.with_columns(
        pl.struct([
            CHARGE_KEY,
            'collisionEnergy',
            INTENSITIES_KEY,
            MZS_KEY,
            PEPTIDE_KEY,
            PROSIT_INTES_KEY,
            PROSIT_IONS_KEY,
            PROSIT_SEQ_KEY,
            PTM_SEQ_KEY,
        ]).apply(
            lambda df_row : calculate_spectral_features(
                df_row,
                ptm_id_weights,
                config.mz_accuracy,
                config.mz_units,
                model,
                str(ox_flag),
                config.delta_method,
                minimal_features=config.minimal_features,
            ),
            skip_nulls=False,
        ).alias('spectralResults')
    )

    spectral_df = spectral_df.filter(pl.col('spectralResults').is_not_null())
    spectral_df = spectral_df.drop(
        [MZS_KEY, INTENSITIES_KEY, PROSIT_INTES_KEY, PROSIT_IONS_KEY]
    )
    spectral_df = spectral_df.unnest('spectralResults')

    spectral_features = deepcopy(SPECTRAL_FEATURES)
    if config.delta_method != 'ignore' and not config.minimal_features:
        spectral_features += DELTA_FEATURES


    spectral_df.write_parquet(
        f'{config.output_folder}/temp_{task_id}_out.parquet'
    )
    del spectral_df
    gc.collect()



""" Script of functions for matching molecular weights to possible fragment ions.
"""
import numpy as np

from inspire.constants import (
    C_TERMINUS,
    N_TERMINUS,
    PROTON,
    RESIDUE_WEIGHTS,
    ION_OFFSET,
)

def compute_potential_mws(sequence, modifications, reverse, ptm_id_weights):
    """ Function to compute the molecular weights of potential fragments
        generated from a peptide (y & b ions, charges 1,2, or 3, and H2O
        or O2 losses).

    Parameters
    ----------
    sequence : str
        The peptide sequence for which we require molecular weights.
    modifications : str
        A string of the ptms for the sequence which will alter
        the potential mzs.
    reverse : bool
        Whether we are getting fragment mzs in the forward direction
        (eg for b ions), or backward direction (eg. for y ions).
    ptm_id_weights : dict
        Mapping of ptm ids to their molecular weights.

    Returns
    -------
    mzs : np.array of floats
        An array of all the possible mzs that coule be observed in
        the MS2 spectrum of a sequence.
    """
    sequence_length = len(sequence)
    n_fragments = sequence_length - 1
    mzs = np.empty(n_fragments)

    if (
        modifications and
        isinstance(modifications, str) and
        modifications != 'nan'
        and modifications != 'None'
    ):
        ptms_list = modifications.split(".")
        mods_list = [int(mod) for mod in ptms_list[1]]

        if reverse:
            ptm_start = int(ptms_list[2])
            ptm_end = int(ptms_list[0])
            mods_list = mods_list[::-1]
        else:
            ptm_start = int(ptms_list[0])
            ptm_end = int(ptms_list[2])
    else:
        mods_list = None
        ptm_start = 0
        ptm_end = 0

    if reverse:
        sequence = sequence[::-1]

    tracking_mw = ptm_id_weights[ptm_start]

    for idx in range(n_fragments):
        tracking_mw += RESIDUE_WEIGHTS[sequence[idx]]
        if mods_list is not None and mods_list[idx]:
            tracking_mw += ptm_id_weights[mods_list[idx]]

        mzs[idx] = tracking_mw

    tracking_mw += RESIDUE_WEIGHTS[sequence[n_fragments]]
    if mods_list is not None and mods_list[n_fragments]:
        tracking_mw += ptm_id_weights[mods_list[n_fragments]]
    if ptm_end:
        tracking_mw += ptm_id_weights[ptm_end]

    return mzs, tracking_mw


def get_ion_masses(sequence, ptm_id_weights, modifications=None):
    """ Function to the get mz's for all the ions predicted by prosit.

    Parameters
    ----------
    sequence : str
        The peptide sequence for which we require molecular weights.
    modifications : str
        A string of the ptms for the sequence which will alter
        the potential mzs.

    Returns
    -------
    all_possible_ions : dict
        A dictionary of all the mzs of all possible b and y ions
        that could be produced.
    """
    sub_seq_mass, total_residue_mass = compute_potential_mws(
        sequence=sequence,
        modifications=modifications,
        reverse=False,
        ptm_id_weights=ptm_id_weights
    )

    rev_sub_seq_mass, _ = compute_potential_mws(
        sequence=sequence,
        modifications=modifications,
        reverse=True,
        ptm_id_weights=ptm_id_weights
    )

    # b- and y-ions for each of the charge options
    prosit_ions = {}
    prosit_ions['b'] = ION_OFFSET['b'] + sub_seq_mass
    prosit_ions['y'] = ION_OFFSET['y'] + rev_sub_seq_mass

    return prosit_ions, (total_residue_mass + C_TERMINUS + N_TERMINUS)


def match_mz(base_mass, frag_z, observed_mzs, loss=0.0):
    """ Function to match a fragment m/z to the nearest experimental m/z.

    Parameters
    ----------
    base_mass : float
        The mass of the b or y ion of that fragment index.
    frag_z : int
        The charge of the fragment ion.
    observed_mzs : np.array of float
        The experimentally observed fragment mzs.
    loss : float (default=0.0)
        The neutral loss weight to be applied.

    Returns
    -------
    mz_error : float
        The mz error on the nearest observed fragment ion.
    matched_mz_ind : int
        The index of the nearest observed fragment ion.
    """
    fragment_mz = (
        (base_mass + (frag_z * PROTON)) - loss
    )/frag_z
    matched_mz_ind = np.argmin(
        np.abs(observed_mzs - fragment_mz)
    )
    return observed_mzs[matched_mz_ind] - fragment_mz, matched_mz_ind



""" Relevant constants used through the inspire pipeline.
"""

# Constants for msp read in.
PROSIT_IONS_KEY = 'prositIons'
PROSIT_INTES_KEY = 'prositIntes'
OXIDATION_PREFIX = 'Oxidation@M'
OXIDATION_PREFIX_LEN = len(OXIDATION_PREFIX)

# Prosit DataFrame column names.
PROSIT_SEQ_KEY = 'modified_sequence'
PROSIT_CHARGE_KEY = 'precursor_charge'
PROSIT_COLLISION_ENERGY_KEY = 'collision_energy'

FIGSHARE_PATH = 'https://figshare.com/ndownloader/articles/20368035/versions/2'
FIGSHARE_EXAMPLE_PATH = 'https://figshare.com/ndownloader/files/37963665'
THERMO_PARSER_PATH = (
    'https://github.com/compomics/ThermoRawFileParser/releases/download/' +
    'v1.4.2/ThermoRawFileParser1.4.2.zip'
)
FIGSHARE_EXTERNAL_UTILS_PATH = 'https://figshare.com/ndownloader/articles/22189198/versions/9'

HEADER_TEXT = '\033[95m'
OKBLUE_TEXT = '\033[94m'
OKCYAN_TEXT = '\033[96m'
OKGREEN_TEXT = '\033[92m'
WARNING_TEXT = '\033[93m'
FAIL_TEXT = '\033[91m'
ENDC_TEXT = '\033[0m'
BOLD_TEXT = '\033[1m'
UNDERLINE_TEXT = '\033[4m'

CHARGE_KEY = 'charge'
PEPTIDE_KEY = 'peptide'
MOD_SEQ_KEY = 'modifiedSequence'
ACCESSION_KEY = 'proteins'
INTENSITIES_KEY = 'intensities'
MZS_KEY = 'mzs'
PTM_SEQ_KEY = 'ptm_seq'
SOURCE_KEY = 'source'
SCAN_KEY = 'scan'
LABEL_KEY = 'Label'
ENGINE_SCORE_KEY = 'engineScore'
SPECTRAL_ANGLE_KEY = 'spectralAngle'
MASS_DIFF_KEY = 'massDiff'
RT_KEY = 'retentionTime'
DELTA_SCORE_KEY = 'deltaScore'
SEQ_LEN_KEY = 'sequenceLength'
ACCESSION_STRATUM_KEY = 'accessionGroup'
SOURCE_INDEX_KEY = 'sourceIndex'

SPEARMAN_KEY = 'spearmanR'
PEARSON_KEY = 'pearsonR'
MAE_KEY = 'medianAbsoluteError'

PTM_NAME_KEY = 'Name'
PTM_ID_KEY = 'Identifier'
PTM_IS_VAR_KEY = 'isVar'
PTM_WEIGHT_KEY = 'Delta'

# Spectral Features:
FRAG_MZ_ERR_MED_KEY = 'medianFragmentMzError'
FRAG_MZ_ERR_VAR_KEY = 'fragmentMzErrorVariance'

MATCHED_IONS_KEY = 'nMatchedIonsDivFrags'
MATCHABLE_IONS_KEY = 'nPrositIonsDivFrags'
NOT_ASSIGNED_KEY = 'nNotMatchableDivFrags'
LOSS_IONS_KEY = 'nLossIonsDivFrags'

MATCHED_INTE_KEY = 'meanMatchedInte'
NON_PRED_INTE_KEY = 'meanPredZeroInte'
MEAN_NOT_MATCHABLE_INTE_KEY = 'meanNotMatchableInte'
LOSS_INTE_KEY = 'meanLossInte'
NOT_MATCHED_INTE_KEY = 'meanNotMatchedInte'

PRECURSOR_INTE_KEY = 'precursorIntensity'

# Percolator column names.
PERC_SCAN_ID = 'scannr'
IN_ACCESSION_KEY = {
    'mokapot': 'proteins',
    'percolator': 'Proteins',
    'percolatorSeparate': 'Proteins',
}
PSM_ID_KEY = {
    'mokapot': 'specID',
    'percolator': 'specID',
    'percolatorSeparate': 'specID',
}
OUT_PSM_ID_KEY = {
    'mokapot': 'specID',
    'percolator': 'PSMId',
    'percolatorSeparate': 'PSMId',
}
OUT_ACCESSION_KEY = {
    'mokapot': 'proteins',
    'percolator': 'proteinIds',
    'percolatorSeparate': 'proteinIds',
}
OUT_SCORE_KEY = {
    'mokapot': 'mokapot score',
    'percolator': 'score',
    'percolatorSeparate': 'score',
}
OUT_Q_KEY = {
    'mokapot': 'mokapot q-value',
    'percolator': 'q-value',
    'percolatorSeparate': 'q-value',
}
OUT_POSTEP_KEY = {
    'mokapot': 'mokapot PEP',
    'percolator': 'posterior_error_prob',
    'percolatorSeparate': 'posterior_error_prob',
}

PREFIX_KEYS = {
    'mokapot': [PSM_ID_KEY['mokapot'], LABEL_KEY, PERC_SCAN_ID],
    'percolator': [PSM_ID_KEY['percolator'], LABEL_KEY, PERC_SCAN_ID],
    'percolatorSeparate': [PSM_ID_KEY['percolatorSeparate'], LABEL_KEY, PERC_SCAN_ID],
}
SUFFIX_KEYS = {
    'mokapot': [PEPTIDE_KEY, IN_ACCESSION_KEY['mokapot']],
    'percolator': [PEPTIDE_KEY, IN_ACCESSION_KEY['percolator']],
    'percolatorSeparate': [PEPTIDE_KEY, IN_ACCESSION_KEY['percolatorSeparate']],
}
FINAL_SCORE_KEY = 'percolatorScore'
FINAL_Q_VALUE_KEY = 'qValue'
FINAL_POSTEP_KEY = 'postErrProb'

PRED_ACCESSION_KEY = 'predictedAccession'
PRED_PEPTIDE_KEY = 'predictedPeptide'
TRUE_ACCESSION_KEY = 'trueAccession'
TRUE_PEPTIDE_KEY = 'truePeptide'

BASIC_FEATURES = [
    MASS_DIFF_KEY,
    SEQ_LEN_KEY,
    ENGINE_SCORE_KEY,
    DELTA_SCORE_KEY,
    CHARGE_KEY,
    'ms1Intensity',
    'nVarMods',
    'missedCleavages',
    'avgResidueMass',
    'nRepeatedResidues',
    'fracUnique',
    'fracKR',
    'fracC',
    'fromChimera',
    'retentionTime',
    'seqLenMeanDiff',
    'absMassDiff',
]

MIN_SEQ_LEN = 7
MAX_SEQ_LEN = 30
MAX_ION_IDX = MAX_SEQ_LEN - 1
MAX_CHARGE = 6
N_LOSSES = 1
N_ION_TYPES = 2
ION_TYPES = 'yb'

MAX_FRAG_CHARGE = 3

PROSIT_PRED_BATCH_SIZE = 1024
PROSIT_ALPHABET = {
    "A": 1,
    "C": 2,
    "D": 3,
    "E": 4,
    "F": 5,
    "G": 6,
    "H": 7,
    "I": 8,
    "K": 9,
    "L": 10,
    "M": 11,
    "N": 12,
    "P": 13,
    "Q": 14,
    "R": 15,
    "S": 16,
    "T": 17,
    "V": 18,
    "W": 19,
    "Y": 20,
    "M(ox)": 21,
}
PROSIT_ALPHABET_S = {integer: char for char, integer in PROSIT_ALPHABET.items()}
PROSIT_UNMOD_ALPHA_S = {
    1: 'A', 2: 'C', 3: 'D',
    4: 'E', 5: 'F', 6: 'G',
    7: 'H', 8: 'I', 9: 'K',
    10: 'L', 11: 'M', 12: 'N',
    13: 'P', 14: 'Q', 15: 'R',
    16: 'S', 17: 'T', 18: 'V',
    19: 'W', 20: 'Y', 21: 'M',
}
PROSIT_MASK_VALUE = -1.0

PROTON = 1.007276466622
ELECTRON = 0.00054858
H = 1.007825035
C = 12.0
O = 15.99491463
N = 14.003074

N_TERMINUS = H
C_TERMINUS = O + H
CO = C + O
CHO = C + H + O
NH2 = N + (H * 2)
H2O = (H * 2) + O
NH3 = N + (H * 3)
LOSS_WEIGHTS = {
    '': 0,
    'NH3': NH3,
    'H2O': H2O
}
NEUTRAL_LOSSES = [
    NH3,
    H2O,
]
LOSS_NAMES = ['',  '*', '&#xb0;']

RESIDUE_WEIGHTS = {
    'A': 71.037114,
    'R': 156.101111,
    'N': 114.042927,
    'D': 115.026943,
    'C': 103.009185,
    'E': 129.042593,
    'Q': 128.058578,
    'G': 57.021464,
    'H': 137.058912,
    'I': 113.084064,
    'L': 113.084064,
    'K': 128.094963,
    'M': 131.040485,
    'F': 147.068414,
    'P': 97.052764,
    'S': 87.032028,
    'T': 101.047679,
    'W': 186.079313,
    'Y': 163.06332,
    'V': 99.068414,
}

KNOWN_PTM_WEIGHTS = {
    'Deamidated (N)': 0.984016,
    'Deamidated (NQ)': 0.984016,
    'Deamidation (NQ)': 0.984016,
    'Deamidation (N)': 0.984016,
    'Deamidation (Q)': 0.984016,
    'Oxidation (M)': 15.994915,
    'Acetyl (N-term)': 42.010565,
    'Acetylation (N-term)': 42.010565,
    'Acetyl (Protein N-term)': 42.010565,
    'Phospho (Y)': 79.966331,
    'Phospho (ST)': 79.966331,
    'Phospho (STY)': 79.966331,
    'Phosphorylation (STY)': 79.966331,
    'Carbamidomethyl (C)': 57.021464,
    'Carbamidomethylation': 57.021464,
    'unknown': 0.0
}

KNOWN_PTM_LOC = {
    'Deamidated (N)': 'N',
    'Deamidated (NQ)': 'NQ',
    'Deamidation (NQ)': 'NQ',
    'Deamidation (N)': 'N',
    'Deamidation (Q)': 'Q',
    'Oxidation (M)': 'M',
    'Acetyl (N-term)': 'N-term',
    'Phospho (Y)': 'Y',
    'Phospho (ST)': 'ST',
    'Phospho (STY)': 'STY',
    'Phosphorylation (STY)': 'STY',
    'Carbamidomethyl (C)': 'C',
    'Carbamidomethylation': 'C',
}

MS2PIP_NAME_MAPPINGS = {
    'Deamidated (N)': 'Deamidation',
    'Deamidated (NQ)': 'Deamidation',
    'Deamidation (NQ)': 'Deamidation',
    'Deamidation (N)': 'Deamidation',
    'Deamidation (Q)': 'Deamidation',
    'Oxidation (M)': 'Oxidation',
    'Acetyl (N-term)': 'Acetyl',
    'Phospho (Y)': 'Phospho',
    'Phospho (ST)': 'Phospho',
    'Phospho (STY)': 'Phospho',
    'Phosphorylation (STY)': 'Phospho',
    'Carbamidomethyl (C)': 'Carbamidomethyl',
    'Carbamidomethylation': 'Carbamidomethyl',
}

ION_OFFSET = {
    'a': N_TERMINUS - CHO,
    'b': N_TERMINUS - H,
    'c': N_TERMINUS + NH2,
    'x': C_TERMINUS + CO - H,
    'y': C_TERMINUS + H,
    'z': C_TERMINUS - NH2,
}

DELTA_PRO_FEATURES = [
    'cTermDist',
    'nTermDist',
    'trueSpectralAngle',
    'inteAtLoc',
    'inteLeft',
    'inteRight',
    MASS_DIFF_KEY,
    'hydroDiff',
    'pkaDiff',
    'polaDiff',
    'BLOSUM6.1',
]

MINIMAL_FEATURE_SET = [
    SPECTRAL_ANGLE_KEY,
    'deltaRT',
    SPEARMAN_KEY,
]

# Constants used by deltapro predictor.

RESIDUE_PROPERTIES = {
    'A': {
        'polarity': 0,
        'hydrophobicity': 1.8,
        'pka': 2.35,
    },
    'C': {
        'polarity': 0,
        'hydrophobicity': 2.5,
        'pka': 1.92,
    },
    'D': {
        'polarity': 1.0,
        'hydrophobicity': -3.5,
        'pka': 1.99,
    },
    'E': {
        'polarity': 1.0,
        'hydrophobicity': -3.5,
        'pka': 2.10,
    },
    'F': {
        'polarity': 0.0,
        'hydrophobicity': -2.8,
        'pka': 2.20,
    },
    'G': {
        'polarity': 0.0,
        'hydrophobicity': -0.4,
        'pka': 2.35,
    },
    'H': {
        'polarity': 1.0,
        'hydrophobicity': -3.2,
        'pka': 1.80,
    },
    'I': {
        'polarity': 0.0,
        'hydrophobicity': 4.5,
        'pka': 2.32,
    },
    'K': {
        'polarity': 1.0,
        'hydrophobicity': -3.9,
        'pka': 2.16,
    },
    'L': {
        'polarity': 0.0,
        'hydrophobicity': 3.8,
        'pka': 2.33,
    },
    'M': {
        'polarity': 0.0,
        'hydrophobicity': 1.9,
        'pka': 2.13,
    },
    'N': {
        'polarity': 1.0,
        'hydrophobicity': -3.5,
        'pka': 2.14,
    },
    'P': {
        'polarity': 0.0,
        'hydrophobicity': -1.6,
        'pka': 1.95,
    },
    'Q': {
        'polarity': 1.0,
        'hydrophobicity': -3.5,
        'pka': 2.17,
    },
    'R': {
        'polarity': 1.0,
        'hydrophobicity': -4.5,
        'pka': 1.82,
    },
    'S': {
        'polarity': 1.0,
        'hydrophobicity': -0.8,
        'pka': 2.19,
    },
    'T': {
        'polarity': 1.0,
        'hydrophobicity': -0.7,
        'pka': 2.09,
    },
    'V': {
        'polarity': 0.0,
        'hydrophobicity': 4.2,
        'pka': 2.29,
    },
    'W': {
        'polarity': 0.0,
        'hydrophobicity': -0.9,
        'pka': 2.46,
    },
    'Y': {
        'polarity': 1.0,
        'hydrophobicity': -1.3,
        'pka': 2.20,
    },
}

BLOSUM6_1_VALUES = {
    'A': 0.19,
    'C': -1.05,
    'D': 0.01,
    'E': -0.08,
    'F': 0.29,
    'G': 1.19,
    'H': -0.79,
    'I': 0.28,
    'K': 0.1,
    'L': 0.34,
    'M': 0.37,
    'N': 0.83,
    'P': -2.02,
    'Q': -0.08,
    'R': 0.2,
    'S': 0.54,
    'T': 0.38,
    'V': 0.16,
    'W': 0.24,
    'Y': -0.48
}

PLOT_AXIS_REQUIREMENTS = {
    'showline':True,
    'linewidth':0.5,
    'linecolor':'black',
    'showgrid':False,
    'ticks':'outside',
}



""" Functions to allow custom logo plotting.
"""
from logomaker import Logo
from logomaker.src.colors import get_color_dict
from logomaker.src.Glyph import Glyph
from logomaker.src.matrix import transform_matrix
from logomaker.src.error_handling import handle_errors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

class CustomLogo(Logo):
    """ Custom class to allow individually set alphas.
    """
    @handle_errors
    def __init__(self,
                 df,
                 color_scheme=None,
                 font_name='sans',
                 stack_order='big_on_top',
                 center_values=False,
                 baseline_width=0.5,
                 flip_below=False,
                 shade_below=0.0,
                 fade_below=0.0,
                 fade_probabilities=False,
                 vpad=0.0,
                 vsep=0.0,
                 alpha=1.0,
                 show_spines=None,
                 ax=None,
                 zorder=0,
                 figsize=(10, 2.5),
                 custom_alphas=None,
                 **kwargs):
        """Modified init function."""
        # set class attributes
        self.df = df
        self.color_scheme = color_scheme
        self.font_name = font_name
        self.stack_order = stack_order
        self.center_values = center_values
        self.baseline_width = baseline_width
        self.flip_below = flip_below
        self.shade_below = shade_below
        self.fade_below = fade_below
        self.fade_probabilities = fade_probabilities
        self.vpad = vpad
        self.vsep = vsep
        self.alpha = alpha
        self.show_spines = show_spines
        self.zorder = zorder
        self.figsize = figsize
        self.ax = ax

        # save other keyword arguments
        self.glyph_kwargs = kwargs

        # register logo as NOT having been drawn
        # This is changed to True after all Glyphs have been rendered
        self.has_been_drawn = False

        # perform input checks to validate attributes
        self._input_checks()

        # compute length
        self.L = len(self.df)

        # get list of characters
        self.cs = np.array([c for c in self.df.columns])
        self.C = len(self.cs)

        # get color dictionary
        # NOTE: this validates color_scheme; not self._input_checks()
        self.rgb_dict = get_color_dict(self.color_scheme, self.cs)

        # get list of positions
        self.ps = np.array([p for p in self.df.index])

        # center matrix if requested
        if self.center_values:
            self.df = transform_matrix(self.df, center_values=True)

        # create axes if not specified by user
        if self.ax is None:
            _, ax = plt.subplots(1, 1, figsize=self.figsize)
            self.ax = ax

        # save figure as attribute
        self.fig = ax.figure

        # compute characters
        if custom_alphas is None:
            self._compute_glyphs()
        else:
            self._compute_custom_alpha_glyphs(custom_alphas)

        # style glyphs below x-axis
        self.style_glyphs_below(shade=self.shade_below,
                                fade=self.fade_below,
                                flip=self.flip_below)

        # fade glyphs by value if requested
        if self.fade_probabilities:
            self.fade_glyphs_in_probability_logo(v_alpha0=0,
                                                 v_alpha1=1)

        # draw
        self.draw()

    
    def _compute_custom_alpha_glyphs(self, custom_alphas):
        """
        Specifies the placement and styling of all glyphs within the logo.
        """
        # Create a dataframe of glyphs
        glyph_df = pd.DataFrame()

        # For each position
        for p in self.ps:

            # get values at this position
            vs = np.array(self.df.loc[p, :])

            # Sort values according to the order in which the user
            # wishes the characters to be stacked
            if self.stack_order == 'big_on_top':
                ordered_indices = np.argsort(vs)

            elif self.stack_order == 'small_on_top':
                tmp_vs = np.zeros(len(vs))
                indices = (vs != 0)
                tmp_vs[indices] = 1.0/vs[indices]
                ordered_indices = np.argsort(tmp_vs)

            elif self.stack_order == 'fixed':
                ordered_indices = np.array(range(len(vs)))[::-1]

            else:
                assert False, 'This line of code should never be called.'

            # Reorder values and characters
            vs = vs[ordered_indices]
            cs = [str(c) for c in self.cs[ordered_indices]]

            # Set floor
            floor = sum((vs - self.vsep) * (vs < 0)) + self.vsep/2.0

            # For each character
            for v, c in zip(vs, cs):

                # Set ceiling
                ceiling = floor + abs(v)

                # Set color
                this_color = self.rgb_dict[c]

                # Set whether to flip character
                flip = (v < 0 and self.flip_below)

                # Create glyph if height is finite
                glyph = Glyph(p, c,
                              ax=self.ax,
                              floor=floor,
                              ceiling=ceiling,
                              color=this_color,
                              flip=flip,
                              zorder=self.zorder,
                              font_name=self.font_name,
                              alpha=custom_alphas[f'{c}{p}'],
                              vpad=self.vpad,
                              **self.glyph_kwargs)

                # Add glyph to glyph_df
                glyph_df.loc[p, c] = glyph

                # Raise floor to current ceiling
                floor = ceiling + self.vsep

        # Set glyph_df attribute
        self.glyph_df = glyph_df
        self.glyph_list = [g for g in self.glyph_df.values.ravel()
                           if isinstance(g, Glyph)]



""" Functions for predicting binding affinity using NetMHCpan
"""
from multiprocessing import Pool
import os

def predict_binding(config):
    """ Function to run binding affinity prediction using NetMHCpan.
    """

    # Run first command so docker image only pulled once.
    if config.pan_docker:
        os.system('docker image pull johncormican/basic-pan-execution')
        alleles_string = ','.join(config.alleles)
        pan_command = config.pan_command
        if pan_command.endswith('/netMHCpan'):
            pan_command = pan_command[:-10]

        os.system(
            f'docker run --rm -v {os.path.abspath(pan_command)}:/net/sund-nas.win.dtu.dk' +
            '/storage/services/www/packages/netMHCpan/4.1/netMHCpan-4.1 ' +
            f'-v {os.path.abspath(config.output_folder)}:/root/output -e ALLELES="{alleles_string}"' +
            f' -e PRED_LIMIT={config.ba_pred_limit} -e N_CORES={config.n_cores} ' +
            'johncormican/basic-pan-execution '
        )

        return

    input_files = [
        in_file for in_file in os.listdir(
            f'{config.output_folder}/mhcpan/'
        ) if in_file.startswith(
            'inputLen'
        )
    ]
    function_args = []
    for allele in config.alleles:
        for input_file in input_files:
            pep_len = int(input_file.split('inputLen')[-1].split('_')[0].split('.')[0])
            if pep_len > config.ba_pred_limit:
                continue
            output_file = input_file.replace(
                f'inputLen{pep_len}', f'output_{pep_len}_{allele}'
            )
            function_args.append(
                f'{config.pan_command} -BA -inptype 1 -a {allele} -l {pep_len} -p -f ' +
                f'{config.output_folder}/mhcpan/{input_file} > ' +
                f'{config.output_folder}/mhcpan/{output_file}'
            )

    with Pool(processes=config.n_cores) as pool:
        pool.map(os.system, function_args)



""" Script containing functions for writing plotting utils of PEPSeek Epitopes results.
"""
from math import ceil
import os

from logomaker import Logo
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from scipy.stats import zscore
import seaborn as sns
from sklearn.decomposition import PCA

from inspire.constants import ENGINE_SCORE_KEY, PLOT_AXIS_REQUIREMENTS, OUT_POSTEP_KEY
from inspire.logo_plot_utils import create_comparison_logo_plot
from inspire.utils import fetch_proteome

def _check_pep(peptide, proteome):
    if '.' in peptide:
        peptide = peptide.split('.')[1]
    for prot in proteome:
        if peptide.replace('I', 'L') in prot[1].replace('I', 'L'):
            return True
    return False

def _check_prot(proteins, proteome):
    proteins = proteins.split(' ')
    for prot in proteome:
        for found_prot in proteins:
            if found_prot == prot[0]:
                return True
    return False

def get_host_peps(config):
    if config.epitope_cut_level == 'psm':
        ep_df = pd.read_csv(f'{config.output_folder}/finalPsmAssignments.csv')
    else:
        ep_df = pd.read_csv(f'{config.output_folder}/finalPeptideAssignments.csv')

    host_prot = fetch_proteome(config.host_proteome, with_desc=False)
    ep_df = ep_df[(ep_df['qValue'] < 0.01) & (ep_df['postErrProb'] < 0.1)]
    ep_df = ep_df[
        (ep_df['peptide'].apply(len) <= 15) & 
        (ep_df['peptide'].apply(len) >= 8)
    ]
    print(ep_df)
    ep_df = ep_df.drop_duplicates(subset=['peptide'])
    ep_df = ep_df[ep_df['peptide'].apply(lambda x : _check_pep(x, host_prot))]
    print(ep_df)

    if config.epitope_cut_level == 'psm':
        host_basic_df = pd.read_csv(f'{config.output_folder}/non_spectral.{config.rescore_method}.psms.txt', sep='\t')
    else:
        host_basic_df = pd.read_csv(f'{config.output_folder}/non_spectral.{config.rescore_method}.peptides.txt', sep='\t')
    
    host_basic_df = host_basic_df[(host_basic_df['q-value'] < 0.01) & (host_basic_df['posterior_error_prob'] < 0.1)]
    host_basic_df = host_basic_df.drop_duplicates(subset=['peptide'])
    host_basic_df = host_basic_df[host_basic_df['peptide'].apply(lambda x : _check_pep(x, host_prot))]
    host_basic_df['peptide'] = host_basic_df['peptide'].apply(lambda x : x.split('.')[1] if '.' in x else x)
    ep_df = pd.merge(ep_df, host_basic_df[['peptide']], how='left', on='peptide', indicator=True)
    ep_df['foundBySearchEngine'] = ep_df['_merge'].apply(lambda x : 'Yes' if x == 'both' else 'No')

    return ep_df

def bar_plot(config, host=False):
    """ Function generate bar plots of shared and PEPseek only identification counts.
    """
    if host:
        host_prot = fetch_proteome(config.host_proteome, with_desc=False)
        host_insp_df = pd.read_csv(f'{config.output_folder}/final.{config.rescore_method}.peptides.txt', sep='\t')
        host_insp_df = host_insp_df[(host_insp_df['q-value'] < 0.01) & (host_insp_df['posterior_error_prob'] < 0.1)]
        host_basic_df = pd.read_csv(f'{config.output_folder}/non_spectral.{config.rescore_method}.peptides.txt', sep='\t')
        host_basic_df = host_basic_df[(host_basic_df['q-value'] < 0.01) & (host_basic_df['posterior_error_prob'] < 0.1)]
        for name in ['peptides', 'proteins']:
            if name == 'peptides':
                joined_df = pd.merge(
                    host_insp_df[['peptide']].drop_duplicates(),
                    host_basic_df[['peptide']].drop_duplicates(),
                    how='outer', indicator=True,
                )
                joined_df = joined_df[joined_df['peptide'].apply(lambda x : _check_pep(x, host_prot))]
                insp_pep_count = joined_df[joined_df['_merge'] == 'left_only'].shape[0]
                shared_pep_count = joined_df[joined_df['_merge'] == 'both'].shape[0]
            else:
                joined_df = pd.merge(
                    host_insp_df[['proteinIds']].drop_duplicates(),
                    host_basic_df[['proteinIds']].drop_duplicates(),
                    how='outer', indicator=True,
                )
                joined_df = joined_df[joined_df['proteinIds'].apply(lambda x : _check_prot(x, host_prot))]
                insp_prot_count = joined_df[joined_df['_merge'] == 'left_only'].shape[0]
                shared_prot_count = joined_df[joined_df['_merge'] == 'both'].shape[0]
    else:
        ep_df = pd.read_csv(f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv')
        insp_df = ep_df[ep_df['foundBySearchEngine'] == 'No']
        shared_df = ep_df[ep_df['foundBySearchEngine'] == 'Yes']

        insp_prot_df = insp_df.drop_duplicates(subset='protein')[['protein']]
        shared_prot_df = shared_df.drop_duplicates(subset='protein')[['protein']]
        shared_pep_count = shared_df.shape[0]
        insp_pep_count = insp_df.shape[0]
        total_df = pd.merge(shared_prot_df, insp_prot_df, how='outer', on='protein', indicator=True)
        insp_prot_count = total_df[total_df['_merge'] == 'right_only'].shape[0]
        shared_prot_count = total_df[total_df['_merge'] != 'right_only'].shape[0]

    fig = px.bar(
        color=['Shared', 'Shared', 'PEPSeek Only', 'PEPSeek Only'],
        y=[shared_pep_count,shared_prot_count,insp_pep_count,insp_prot_count],
        x=['Epitope', 'Antigen', 'Epitope', 'Antigen',],
        color_discrete_map={
            'PEPSeek Only': '#FFBE00',
            'Shared': '#C0C0C0',
        },
    )
    fig.update_traces(marker_line_color='black', marker_line_width=0.5)
    fig.update_xaxes(PLOT_AXIS_REQUIREMENTS, dtick=1, title='')
    fig.update_yaxes(PLOT_AXIS_REQUIREMENTS, title='Number of Epitopes/Antigens',)
    fig.update_layout(
        barmode='group',
        font_family='Helvetica',
        font_color='black',
        width=500,
        height=300,
        title_x=0.5,
        plot_bgcolor='rgba(0,0,0,0)',
        yaxis_showticklabels=True,
    )
    code = ''
    if host:
        code = '_host'
    fig.write_image(
        f'{config.output_folder}/img/PEPSeek{code}_bar_plot.svg', engine='kaleido'
    )


def check_engine_results(pathogen_df, config):
    """ Function to check if identified PSMs were found using the original search engine.

    Parameters
    ----------
    pathogen_df : pd.DataFrame
        DataFrame of identified pathogen peptides.
    config : inspire.config.Config
        The Config object which controls the experiment.

    Returns
    -------
    pathogen_df : pd.DataFrame
        The input DataFrame with information added on the original search engine identifications.
    """
    pathogen_df['maxEngineScore'] = pathogen_df.groupby(
        'peptide'
    )[ENGINE_SCORE_KEY].transform('max')

    if config.engine_score_cut is not None:
        pathogen_df['foundBySearchEngine'] = pathogen_df['maxEngineScore'].apply(
            lambda x : 'Yes' if x > config.engine_score_cut else 'No'
        )
    else:
        if config.epitope_cut_level == 'psm':
            non_spect_df = pd.read_csv(
                f'{config.output_folder}/non_spectral.{config.rescore_method}.psms.txt',
                sep='\t',
            )
        else:
            non_spect_df = pd.read_csv(
                f'{config.output_folder}/non_spectral.{config.rescore_method}.peptides.txt',
                sep='\t',
            )
        non_spect_df = non_spect_df[
            (non_spect_df[OUT_POSTEP_KEY[config.rescore_method]] < config.epitope_candidate_cut_off)
        ]
        non_spect_df = non_spect_df.drop_duplicates(subset='peptide')
        non_spect_df = non_spect_df[['peptide']]
        non_spect_df['peptide'] = non_spect_df['peptide'].apply(
            lambda peptide : peptide.split('.')[1] if '.' in peptide else peptide
        )
        non_spect_df['foundBySearchEngine'] = 'Yes'

        pathogen_df = pd.merge(pathogen_df, non_spect_df, how='left', on='peptide')
        pathogen_df['foundBySearchEngine'] = pathogen_df['foundBySearchEngine'].apply(
            lambda x : x if x == 'Yes' else 'No'
        )

    pathogen_df = pathogen_df.drop('maxEngineScore', axis=1)

    return pathogen_df

COLOUR_DICTS = {
    'No': '#FFBE00',
    'Yes': 'darkgrey',
}

def swarm_plots(config, host=False):
    """ Function to provide bee swarm plots of various metrics for shared and PEPSeek
        only identifications. 
    """
    if host:
        ep_df = get_host_peps(config)
    else:
        ep_df = pd.read_csv(f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv')
    fig = make_subplots(rows=1, cols=3)
    range_cuts = []
    for met_idx, metric in enumerate(['spectralAngle', 'engineScore', 'deltaRT']):
        if host:
            for found_val in ['No', 'Yes']:
                sub_df = ep_df[ep_df['foundBySearchEngine'] == found_val]
                fig.add_trace(go.Violin(
                    fillcolor=COLOUR_DICTS[found_val],
                    y=sub_df[metric],
                    x=sub_df[
                        'foundBySearchEngine'
                    ].str.replace('Yes', 'Shared').replace('No', 'PEPSeek Only'),
                    line_color='black', line_width=0.5,
                    points=False,
                    opacity=0.8,
                    meanline_visible=True,
                ), row=1, col=1+met_idx)
        else:
            mini_fig = px.strip(
                color=ep_df['foundBySearchEngine'],
                y=ep_df[metric],
                x=ep_df[
                    'foundBySearchEngine'
                ].str.replace('Yes', 'Shared').replace('No', 'PEPSeek Only'),
                color_discrete_map={
                    'No': '#FFBE00',
                    'Yes': 'darkgrey',
                },
            )
            for trace in mini_fig['data']:
                fig.add_trace(trace, row=1, col=1+met_idx)
        range_cuts.append(
            10 + 10*(ep_df[metric].max()//10)
        )

    fig.update_traces(marker_line_color='black', marker_line_width=0.5)
    fig.update_xaxes(PLOT_AXIS_REQUIREMENTS, dtick=1)
    fig.update_yaxes(PLOT_AXIS_REQUIREMENTS)
    fig.update_layout(
        {
            'yaxis1': {'title': 'Spectral Angle', 'range':[0,1]},
            'yaxis2': {'title': 'Search Engine Score', 'range': [0,range_cuts[1]]},
            'yaxis3': {'title': 'Retention Time Error', 'range': [0,range_cuts[2]]},
        },
        barmode='group',
        font_family='Helvetica',
        font_color='black',
        width=900,
        height=400,
        title_x=0.5,
        plot_bgcolor='rgba(0,0,0,0)',
        yaxis_showticklabels=True,
        showlegend=False,
    )
    code = ''
    if host:
        code = '_host'
    fig.write_image(
        f'{config.output_folder}/img/PEPSeek{code}_metrics.svg', engine='kaleido'
    )


def plot_binding_clustermap(config, host=False):
    """ Function to plot clustermap of peptides and their binding affinities.
    """
    if host:
        ep_df = get_host_peps(config)
    else:
        ep_df = pd.read_csv(f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv')

    ep_df.index = ep_df['peptide']
    binding_aff_cols = [col for col in ep_df.columns if col.endswith('%Rank_BA')]
    if len(binding_aff_cols) < 2:
        return
    ep_df = ep_df[binding_aff_cols]
    ep_df = ep_df.rename(
        columns={col:col.strip('_%Rank_BA') for col in ep_df.columns}
    )
    ep_df = ep_df.map(lambda x : round(x, 2))
    ep_df = ep_df.sort_values(by=list(ep_df.columns))

    if host:
        cluster_map = sns.clustermap(
            ep_df, cmap=sns.color_palette("blend:#FA3F37,#FAF4D3", as_cmap=True),
            yticklabels=False, figsize=(7,10), vmin=0, vmax=20,)
    else:
        fig_length = 1+(ep_df.shape[0]//6)

        cluster_map = sns.clustermap(
            ep_df, cmap=sns.color_palette("blend:#FA3F37,#FAF4D3", as_cmap=True),
            yticklabels=True, figsize=(7,fig_length), vmin=0, vmax=20,
            linewidths=0.5, linecolor='black')

    code = ''
    if host:
        code = '_host'
    cluster_map.figure.savefig(f'{config.output_folder}/img/PEPSeek{code}_affinity_cluster.svg')

def plot_quant_pca(config):
    """ Function to plot PCA first two components of quantification across files for pathogen
        peptides and a random selection of host peptides.
    """
    if not os.path.exists(f'{config.output_folder}/quant/quantified_per_file.csv'):
        return
    ep_df = pd.read_csv(f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv')
    quant_df = pd.read_csv(f'{config.output_folder}/quant/quantified_per_file.csv')
    quant_idp_cols = [col for col in quant_df.columns if col.endswith('_idp')]
    quant_raw_cols = [col for col in quant_df.columns if col.endswith('_raw')]
    quant_df = quant_df[['peptide'] + quant_raw_cols + quant_idp_cols]
    z_score_df = pd.DataFrame(
        zscore(quant_df[quant_raw_cols], axis=1, nan_policy='omit'),
        index=quant_df.index,
        columns=quant_raw_cols,
    )
    quant_df = quant_df.drop(quant_raw_cols, axis=1)
    quant_df = pd.merge(quant_df, z_score_df, left_index=True, right_index=True)

    merged_df = pd.merge(ep_df[['peptide']], quant_df, how='inner', on='peptide')
    host_samples = 100*(len(merged_df)//100) + 100

    quant_df = pd.merge(quant_df, merged_df[['peptide']], how='left', on='peptide', indicator=True)
    quant_df = quant_df[quant_df['_merge'] == 'left_only'].drop('_merge', axis=1).sample(
        n=host_samples,
        random_state=42,
    )

    merged_df['group'] = 'Pathogen'
    quant_df['group'] = 'Host'

    total_df = pd.concat([merged_df, quant_df])
    total_df = total_df.dropna(axis=0)

    pca = PCA(n_components=2).fit(total_df[quant_raw_cols+quant_idp_cols])

    component_1_key = f'Component 1 (Exp. Var. {round(pca.explained_variance_ratio_[0], 2)})'
    component_2_key = f'Component 2 (Exp. Var. {round(pca.explained_variance_ratio_[1], 2)})'

    pca_df = pd.DataFrame(
        pca.transform(total_df[quant_raw_cols+quant_idp_cols]),
        columns=[
            component_1_key, component_2_key,
        ],
        index=total_df.index
    )

    pca_df /= pca_df.abs().max().max()
    pca_df['Accession'] = total_df['group']

    fig = px.scatter(
        pca_df,
        x=component_2_key,
        y=component_1_key,
        color='Accession',
        color_discrete_map={'Host': 'darkgreen', 'Pathogen': '#B65FCF'},
    )
    fig.update_traces(opacity=1)
    fig.update_traces(marker_line_color='black', marker_line_width=0.5)
    fig.update_traces(marker_size=8)
    fig.update_xaxes(PLOT_AXIS_REQUIREMENTS, range=[-1.5,1.5])
    fig.update_yaxes(PLOT_AXIS_REQUIREMENTS, range=[-1.5,1.5])
    fig.update_layout(
        barmode='group',
        font_family='Helvetica',
        font_color='black',
        width=600,
        height=500,
        title_x=0.5,
        plot_bgcolor='rgba(0,0,0,0)',
    )
    fig.write_image(
        f'{config.output_folder}/img/PEPSeek_pca.svg', engine='kaleido'
    )


def js_divergence(config, host_df):
    """ Function plot JS divergence between pathogen and host 9-mer peptides.
    """
    ep_df = pd.read_csv(f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv')
    ep_df = ep_df[['peptide']]
    ep_df = ep_df[ep_df['peptide'].apply(lambda x : len(x) == 9)]
    host_df = host_df[['peptide']].drop_duplicates()
    host_df = host_df[host_df['peptide'].apply(lambda x : len(x) == 9)]
    
    create_comparison_logo_plot(
        [ep_df, host_df], ['Pathogen', 'Host'], 9,
        f'{config.output_folder}/img', 'logo_comp_plots'
    )



""" Functions for reading in Mascot search results.
"""
import re

import pandas as pd
import polars as pl

from inspire.constants import (
    ACCESSION_KEY,
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    ENGINE_SCORE_KEY,
    LABEL_KEY,
    MASS_DIFF_KEY,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_IS_VAR_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    RT_KEY,
    SCAN_KEY,
    SEQ_LEN_KEY,
    SOURCE_KEY,
)
from inspire.utils import filter_for_prosit

# Define separators within Mascot output and the names for relevant columns.
MASCOT_HEADER_MARKER = 'Header'
MASCOT_FILENAME_MARKER = 'Peak list data path'
MASCOT_HITS_START_MARKER = 'prot_hit_num'
MASCOT_NO_FIXED_MODS_MARKER = '""'
MASCOT_QUERIES_START_MARKER = 'Queries'
MASCOT_VAR_MODS_MARKER = 'Variable modifications'
MASCOT_FIXED_MODS_MARKER = 'Fixed modifications'
MASCOT_SEARCH_PARAMS_MARKER = 'Search Parameters'
MASCOT_SCAN_FILE_NAME_LINE = 'Peak list data path'

MASCOT_CHARGE_KEY = 'pep_exp_z'
MASCOT_MISS_KEY = 'pep_miss'
MASCOT_DECOY_KEY = 'prot_desc'
MASCOT_ENGINE_SCORE_KEY = 'pep_score'
MASCOT_PEPTIDE_KEY = 'pep_seq'
MASCOT_PEP_QUERY_KEY = 'pep_query'
MASCOT_SCAN_TITLE_KEY = 'pep_scan_title'
MASCOT_PTM_SEQ_KEY = 'pep_var_mod_pos'
MASCOT_ACCESSION_KEY = 'prot_acc'
MASCOT_MASS_KEY = 'pep_exp_mr'
MASCOT_PRED_MASS_KEY = 'pep_calc_mr'
REQUIRED_MASCOT_COLUMNS = [
    MASCOT_PEP_QUERY_KEY,
    MASCOT_ACCESSION_KEY,
    MASCOT_CHARGE_KEY,
    MASCOT_DECOY_KEY,
    MASCOT_ENGINE_SCORE_KEY,
    MASCOT_MASS_KEY,
    MASCOT_MISS_KEY,
    MASCOT_PEPTIDE_KEY,
    MASCOT_PRED_MASS_KEY,
    MASCOT_PTM_SEQ_KEY,
    MASCOT_SCAN_TITLE_KEY,
]
MASCOT_QUERIES_TITLE_KEY = 'StringTitle'
MASCOT_QUERIES_RT_KEY = 'Retention time range'
MASCOT_QUERIES_INTENSITY_KEY = 'intensity'

def _get_mascot_file_metadata(csv_file):
    """ Function to extract metadata from a csv file of Mascot results.

    Parameters
    ----------
    csv_file : str
        The input address for a mascot search results file.

    Returns
    -------
    header_line : int
        The line number of the Header line.
    output_filename : str
        The name to be used for the output file of processing to ensure
        it matches with the processed Peaks file.
    hits_line : int
        The line number where the data on peptide matches starts.
    queries_line : int
        The line number where the query data starts (including Retention
        Time data).
    """
    with open(csv_file, 'r', encoding='UTF-8') as open_file:
        line_idx = 0
        header_line = -1
        hits_line = -1
        queries_line = None
        fixed_mods_line = None
        var_mods_line = -1
        search_params_line = -1
        scan_filename = ''

        while (line := open_file.readline()):
            if header_line==-1 and MASCOT_HEADER_MARKER in line:
                header_line = line_idx
            if (
                fixed_mods_line is None and
                MASCOT_FIXED_MODS_MARKER in line and
                MASCOT_NO_FIXED_MODS_MARKER not in line
            ):
                fixed_mods_line = line_idx
            if var_mods_line == -1 and MASCOT_VAR_MODS_MARKER in line:
                var_mods_line = line_idx
            if MASCOT_SCAN_FILE_NAME_LINE in line:
                scan_filename = line.split(',')[-1].strip('\n').strip('"')
            if search_params_line == -1 and MASCOT_SEARCH_PARAMS_MARKER in line:
                search_params_line = line_idx
            if hits_line==-1 and MASCOT_HITS_START_MARKER in line:
                hits_line = line_idx
            if queries_line is None and MASCOT_QUERIES_START_MARKER in line:
                queries_line = line_idx
                break
            line_idx += 1

    mod_ranges = {}
    mod_ranges['variable'] = range(var_mods_line+2, search_params_line-1)
    if fixed_mods_line is not None:
        mod_ranges['fixed'] = range(fixed_mods_line+2, var_mods_line-1)

    return header_line, hits_line, queries_line, mod_ranges, scan_filename

def _skip_logic(idx, start_idx=0, end_idx=None):
    """ Simple function for identifying which lines to skip when reading a csv.

    Parameters
    ----------
    idx : int
        A line number in a file.
    start_idx : int
        The line number at which csv data starts.
    end_idx: int
        The line number at which csv data ends.

    Returns
    ------
    skip_line : bool
        A flag on whether to skip the line or not.
    """
    if end_idx is not None:
        if start_idx <= idx < end_idx:
            return False
    else:
        if idx >= start_idx:
            return False
    return True

def separate_scan_and_source(scan_title, scan_title_format, source_list=None, source_filename=None):
    """ Function to separate source file and scan number (as well as retention time if
        Distiller format used) from mascots scan_title_format column.

    Parameters
    ----------
    df_row : pd.Series
        A row of the mascot search results DataFrame.
    scan_title_format : str or None
        The format of the scan title (see README for options).
    source_list : str or None
        A list of the source files if Mascot Distiller was used.
    source_filename : str or None
        The name of the source file used in the search if available.

    Returns
    -------
    df_row : pd.Series
        The input row updated with new columns.
    """
    results = {}
    if scan_title_format is None:
        results[SCAN_KEY] = int(scan_title.split('=')[-1].strip('~'))
        if source_filename is None:
            source = scan_title.split('File:')[-1]
            if source.startswith('~'):
                source = source.split('~')[1]
            else:
                source = source.split(', ')[0]
        else:
            source = source_filename

        if source.endswith('.raw') or source.endswith('.mgf'):
            source = source[:-4]
        results[SOURCE_KEY] = source

    elif scan_title_format == 'mascotDistiller':
        scan_rt_details = scan_title.split(' Scan ')[-1].split(' (rt')
        results[SCAN_KEY] = int(scan_rt_details[0])
        source_idx = int(scan_title.split(' from file [')[-1].strip(']'))
        source_idx = int(scan_title.split(' from file [')[-1].strip(']'))
        results[SOURCE_KEY] = source_list[source_idx]
        results[RT_KEY] = float(scan_rt_details[-1][1:].split(')')[0])
    elif scan_title_format == 'distiller':
        split_name = scan_title.split(' (rt=')
        results[SCAN_KEY] = int(split_name[0].split(' Scan ')[-1])
        results[RT_KEY] = float(split_name[-1].split(')')[0])
    else:
        raise ValueError('Oops')
    return results

def _read_mascot_dfs(
        csv_filename,
        hits_line,
        queries_line,
        mods_range,
        scan_file,
    ):
    hits_df = pd.read_csv(
        csv_filename,
        skiprows=lambda idx : _skip_logic(idx, hits_line, queries_line),
        usecols=REQUIRED_MASCOT_COLUMNS
    )
    hits_df = pl.from_pandas(hits_df)

    mods_df = pd.read_csv(
        csv_filename,
        skiprows=lambda idx : idx not in mods_range['variable'],
        usecols=range(4),
    )
    mods_df[PTM_IS_VAR_KEY] = True

    # Rename to match inSPIRE naming scheme.
    hits_df = hits_df.rename({
        MASCOT_ACCESSION_KEY: ACCESSION_KEY,
        MASCOT_CHARGE_KEY: CHARGE_KEY,
        MASCOT_ENGINE_SCORE_KEY: ENGINE_SCORE_KEY,
        MASCOT_PTM_SEQ_KEY: PTM_SEQ_KEY,
        MASCOT_PEPTIDE_KEY: PEPTIDE_KEY,
        MASCOT_MISS_KEY: 'missedCleavages',
    })

    # Filter for Prosit and add feature columns not present.
    hits_df = filter_for_prosit(hits_df)
    hits_df = hits_df.with_columns(
        (pl.col(MASCOT_MASS_KEY) - pl.col(MASCOT_PRED_MASS_KEY)).alias(MASS_DIFF_KEY),
        pl.col(PEPTIDE_KEY).str.lengths().alias(SEQ_LEN_KEY),
    )
    hits_df = hits_df.with_columns(
        (pl.col(MASCOT_MASS_KEY)/pl.col(SEQ_LEN_KEY)).alias('avgResidueMass'),
    )

    hits_df = hits_df.drop([MASCOT_MASS_KEY, MASCOT_PRED_MASS_KEY])

    hits_df = hits_df.with_columns(
        pl.col(MASCOT_PEP_QUERY_KEY).apply(
            lambda x : scan_file + str(x)
        ).alias(MASCOT_PEP_QUERY_KEY),
        pl.col(MASCOT_DECOY_KEY).apply(
            lambda x : -1 if 'Reversed' in x or 'random' in x or 'Random' in x else 1
        ).alias(LABEL_KEY)
    )

    hits_df = hits_df.drop([MASCOT_DECOY_KEY])

    # Add fixed mods
    if 'fixed' in mods_range:
        fixed_mods_df = pd.read_csv(
            csv_filename,
            skiprows=lambda idx : idx not in mods_range['fixed'],
            usecols=range(4)
        )
        fixed_mods_df[PTM_ID_KEY] = fixed_mods_df[PTM_ID_KEY] + mods_df.shape[0]
        fixed_ptm_dict = {}
        for mod_name, mod_id in zip(
                fixed_mods_df[PTM_NAME_KEY].tolist(),
                fixed_mods_df[PTM_ID_KEY].tolist()
            ):
            modified_residues = mod_name.split('(')[-1].split(')')[0]
            fixed_ptm_dict[modified_residues] = mod_id

        hits_df[PTM_SEQ_KEY] = hits_df[[PEPTIDE_KEY, PTM_SEQ_KEY]].apply(
            lambda x : _add_fixed_mod(x[PEPTIDE_KEY], x[PTM_SEQ_KEY], fixed_ptm_dict),
            axis=1
        )

        fixed_mods_df[PTM_IS_VAR_KEY] = False

        mods_df = pd.concat([mods_df, fixed_mods_df])


    return hits_df, mods_df

def _add_fixed_mod(peptide, ptm_seq, fixed_ptm_dict):
    """ Function to add fixed modifications to Mascot search results in the same style as
        variable modifications.

    Parameters
    ----------
    peptide : str
        The peptide sequence.
    ptm_seq : str
        The sequence of variable modifications.
    fixed_ptm_dict : dict
        A dictionary mapping fixed ptm residues to their ID.

    Returns
    -------
    ptm_seq : str
        The input ptm_seq updated with any relevant fixed modifications.
    """
    for residues in fixed_ptm_dict:
        if residues == 'N-term':
            if not isinstance(ptm_seq, str):
                ptm_seq = '0.' + ''.join(['0']*len(peptide)) + '.0'
            ptm_seq = str(fixed_ptm_dict[residues]) + ptm_seq[1:]
        elif residues == 'C-term':
            if not isinstance(ptm_seq, str):
                ptm_seq = '0.' + ''.join(['0']*len(peptide)) + '.0'
            ptm_seq = ptm_seq[:-1] + str(fixed_ptm_dict[residues])
        else:
            for res in residues:
                ptm_postns = [m.start() for m in re.finditer(res, peptide)]
                if ptm_postns and not isinstance(ptm_seq, str):
                    ptm_seq = '0.' + ''.join(['0']*len(peptide)) + '.0'
                for pos in ptm_postns:
                    ptm_seq = ptm_seq[:pos+2] + str(fixed_ptm_dict[residues]) + ptm_seq[pos+3:]

    return ptm_seq

def read_single_mascot_data(input_filename, scan_title_format, variable_mods):
    """ Function to read in mascot search results from a single file.

    Parameters
    ----------
    input_filename : str
        A location of mascot search results.
    scan_title_format : str
        The format of mascot's pep_scan_title column.
    source_list : list of str or None
        A list of the source files used in the mascot search.
    variable_mods : pd.DataFrame
        The previously discovered variable modifications.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    # Original File contains both peptide hits and queries, search file to find separators.
    header_line, hits_line, queries_line, mod_ranges, scan_file = _get_mascot_file_metadata(
        input_filename
    )

    if header_line == -1:
        raise ValueError('No header line found in Mascot search results.')

    # Read separate files as pandas dataframes.
    hits_df, new_variable_mods = _read_mascot_dfs(
        input_filename,
        hits_line,
        queries_line,
        mod_ranges,
        scan_file,
    )

    if scan_title_format == 'distiller':
        hits_df = hits_df.with_columns(
            pl.lit(scan_file.strip('.raw').strip('.mgf').strip('.temp')).alias(SOURCE_KEY)
        )

    if variable_mods is None or new_variable_mods.equals(variable_mods):
        return hits_df, new_variable_mods

    new_variable_mods = new_variable_mods.rename(
        columns={
            PTM_ID_KEY: 'newId'
        }
    )[[PTM_NAME_KEY, 'newId']]
    combo_df = pd.merge(
        variable_mods,
        new_variable_mods,
        on=[PTM_NAME_KEY],
        how='outer',
    )
    unmatched_df = combo_df[combo_df['newId'] != combo_df[PTM_ID_KEY]]
    replacements = {str(k): str(v) for k, v in zip(
        unmatched_df['newId'].tolist(), unmatched_df[PTM_ID_KEY].tolist()
    )}
    hits_df = hits_df.with_columns(
        pl.col(PTM_SEQ_KEY).apply(
            lambda x : _replace_ptm_id(x, replacements)
        )
    )

    return hits_df, variable_mods

def _replace_ptm_id(ptm_seq, replacements):
    """ Function to replace certain ptm ids in the ptm seq.

    Parameters
    ----------
    ptm_seq : str or None
        The PTM sequence for the PSM.
    replacements : dict
        The PTM IDs to be updated.

    Returns
    -------
    new_ptm_seq : str or None
        The updated PTM sequence.
    """
    if not isinstance(ptm_seq, str):
        return ptm_seq
    new_ptm_seq = ''
    for char in ptm_seq:
        if char in replacements:
            new_ptm_seq += replacements[char]
        else:
            new_ptm_seq += char
    return new_ptm_seq

def read_mascot_data(
        mascot_data,
        scan_title_format,
        source_list,
        reduce,
        source_filename,
        with_accession=False,
    ):
    """ Function to read in mascot search results from one or more files.

    Parameters
    ----------
    mascot_data : str or list of str
        A single location of mascot search results or a list of locations.
    scan_title_format : str
        The format of mascot's pep_scan_title column.
    source_list : list of str or None
        A list of the source files used in the mascot search.
    source_filename : str or None
        The name of the source file used in the search if available.

    Returns
    -------
    hits_df : pl.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    if isinstance(mascot_data, list):
        hits_dfs = []
        variable_mods_dfs = []
        variable_mods = None
        for input_filename in mascot_data:
            hits_df, variable_mods = read_single_mascot_data(
                input_filename,
                scan_title_format,
                variable_mods,
            )
            hits_dfs.append(hits_df)
            variable_mods_dfs.append(variable_mods)
        hits_df = pl.concat(hits_dfs)
        for i in range(len(variable_mods_dfs)-1):
            assert variable_mods_dfs[i].equals(variable_mods_dfs[i+1])
    else:
        hits_df, variable_mods = read_single_mascot_data(
            mascot_data,
            scan_title_format,
            None,
        )

    if with_accession:
        hits_df = hits_df.with_columns(
            pl.col(ACCESSION_KEY).apply(
                lambda x : 1 if 'PSP' in x else 0
            ).alias('PSP')
        )
        hits_df = hits_df.sort(by='PSP')
        hits_df = hits_df.sort(by=LABEL_KEY, descending=True)

    hits_df = mascot_reduce_to_max(hits_df, reduce, with_accession)

    hits_df = hits_df.with_columns(
        pl.col(MASCOT_SCAN_TITLE_KEY).apply(
            lambda x : separate_scan_and_source(x, scan_title_format, source_list, source_filename),
            skip_nulls=False,
        ).alias('results')
    ).unnest('results')
    hits_df = hits_df.unique(
        subset=[SOURCE_KEY, SCAN_KEY, PEPTIDE_KEY]
    )

    hits_df = hits_df.with_columns(
        pl.lit(0).alias('fromChimera')
    )

    hits_df = hits_df.drop(MASCOT_SCAN_TITLE_KEY)

    return hits_df, variable_mods

def mascot_reduce_to_max(main_df, reduce, with_accession):
    """ Function to filter down to the best scoring PSM per spectrum.

    Parameters
    ----------
    main_df : pl.DataFrame
        A DataFrame of search results.

    Returns
    -------
    main_df : pl.DataFrame
        The search DataFrame with results filtered per PSM.
    """
    if with_accession:
        main_df = main_df.sort(
            by=[LABEL_KEY, 'PSP', ENGINE_SCORE_KEY, PEPTIDE_KEY],
            descending=[True, False, True, True],
        )
    else:
        main_df = main_df.sort(
            by=[LABEL_KEY, ENGINE_SCORE_KEY, PEPTIDE_KEY], descending=True
        )
    if reduce:
        main_df = main_df.with_columns(
            pl.col(PEPTIDE_KEY).apply(
                lambda x : x.replace('I', 'L')
            ).alias('ilSub')
        )
        main_df = main_df.unique(subset=[MASCOT_PEP_QUERY_KEY, 'ilSub'])
        main_df = main_df.drop('ilSub')

    main_df = main_df.with_columns(
        pl.lit(0).alias(DELTA_SCORE_KEY)
    )

    return main_df
